/**

	Valentin Tihomirov, 2006, www.abelectron.com

Wraps the stream being generated into a stream.

To enable java comm: 
	1. get javacomm20-win32.zip.
	2. Copy  
	
FILE                  |TARGET DIRECTORY    | This provides the Java with
----------------------+--------------------+----------------------------
win32com.dll		  |jre_ver\bin         | a windows serial driver
----------------------+--------------------+----------------------------
javax.comm.properties |jre_ver\lib         | a path to win32comm.dll
----------------------+--------------------+----------------------------
comm.jar			  |jre_ver\lib\ext     | communication classes
----------------------+--------------------+----------------------------
comm.jar			  |jdk_ver\jre\lib\ext | optionall comm classes for compiler

*/

import java.io.*;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.comm.*;


import gnu.getopt.Getopt;


public class FluentOscillator {

	public static void main(String[] args) throws Exception {
		/*int i = 128;
		byte b1 = (byte) i;
		log((((int) b1) & 0xFF) + "");
		System.exit(0);*/
		
		final int	BUFFER_SIZE = 128000;
		byte nChannels = 2; // actually we need mono but on one of my audio cards wierdly bubbles in the mono mode
		
		PeriodicSignal oscillator = new Sine(); // play sin by default
		float	sampleRate = 44100.0F;
		float	signalFrequency = 1000.0F;
		float	amplitude = 0.7F;
		int		sampleSizeInBits = 16; // 8 or 16
		String outputStreamType = "audio", outputStreamName = null;
		AudioFormat.Encoding enc = AudioFormat.Encoding.PCM_SIGNED;
		float duration = 0; // infinity

				
		//
		// Parse arguments
		//
		
			
		int	c;
		if (args.length == 0)
			log("use -? option for usage");
		else {
			Getopt opt = new Getopt("AudioPlayer", args, "?umo:r:f:a:b:d:p:n:");
			while ((c = opt.getopt()) != -1) {
				switch (c) {
					case '?': printUsageAndExit(); break;
					case 'm': nChannels = 1; break;
					case 'u': enc = AudioFormat.Encoding.PCM_UNSIGNED; break;
					case 'o': oscillator = (PeriodicSignal) Class.forName(opt.getOptarg()).newInstance(); break;
					case 'r': sampleRate = Float.parseFloat(opt.getOptarg()); break;
					case 'f': signalFrequency = Float.parseFloat(opt.getOptarg()); break;
					case 'a': amplitude = Float.parseFloat(opt.getOptarg()); break;
					case 'p': outputStreamType = opt.getOptarg(); break;
					case 'n': outputStreamName = opt.getOptarg(); break;
					case 'b': sampleSizeInBits = Integer.parseInt(opt.getOptarg()); break;
					case 'd': duration = Float.parseFloat(opt.getOptarg()); break;
					default: break;
				}
			}
		}
		
		long remainingFrames = remainingFrames = (duration == 0) ? AudioSystem.NOT_SPECIFIED : (long) (duration * sampleRate);
		
		// 
		// Instantiate input
		//
		
		final int frameSize = (int) Math.ceil((float)(sampleSizeInBits * nChannels)/8);
		final AudioFormat audioFormat = new AudioFormat(enc, sampleRate, sampleSizeInBits, nChannels, frameSize, sampleRate, false);
/*		OscillatorInputStream in = enc == AudioFormat.Encoding.PCM_UNSIGNED ? 
			new IdealOscillatorInputStream(oscillator, amplitude, signalFrequency, audioFormat, remainingFrames) :
			new OscillatorInputStream(oscillator, amplitude, signalFrequency, audioFormat, remainingFrames);*/
			
		OscillatorInputStream in = OscillatorInputStreamFactory.get(oscillator, amplitude, signalFrequency, audioFormat, remainingFrames);
		try {
			
			
			//
			// Instantiate output
			//
			
			final OutputStream out;
			if (outputStreamType.compareTo("audio") == 0)
				out = new LineOutputStream(audioFormat);
			else if (outputStreamType.compareTo("file") == 0)
				if (outputStreamName == null)
					throw new IllegalArgumentException("Output file name must be specified");
				else
					out = new FileOutputStream(outputStreamName);
			else if (outputStreamType.compareTo("ser") == 0) {
				if (outputStreamName == null) {
					log("Shoing ports");
					java.util.Enumeration portList = CommPortIdentifier.getPortIdentifiers();
					while (portList.hasMoreElements()) {
						CommPortIdentifier portId = (CommPortIdentifier) portList.nextElement();
						log(portId.getName() + " " + (portId.getPortType() == CommPortIdentifier.PORT_SERIAL ? "ser" : "par"));
					}
					return;
				}
				
				out = CommPortIdentifier.getPortIdentifier(outputStreamName).open("what the hell id they need?", 2000).getOutputStream();

			} else if (outputStreamType.compareTo("console") == 0)
				out = new ConsoleOutputStream();
			else
				throw new IllegalArgumentException ("Unsupported output type - " + outputStreamType);
			
				
			try {
				
				//
				// Play Loop
				//
				
				byte[] buf = new byte[BUFFER_SIZE];
			
				log ("press Contorl-C to terminate");
				
				int nBytesRead;
				while ((nBytesRead = in.read(buf)) != -1)
					out.write(buf, 0, nBytesRead);
					
			} finally {
				out.close();
			}
			
		} finally {
			in.close();
		}
	}



	static void printUsageAndExit() {
		log("usage: java OscillatorPlayer [-m] [-u] [-o <waveformtype>] [-a <amplitude>] [-f <signalfrequency>] [-r <samplerate>] [-b <8/16 bit precision>] [-d <duration>] [-p <file|ser|audio|console> -n <output name>]");
		log("\tamplitude: 0..1");
		log("\tm = mono, u = unsigned PCM");
		log("\toscillator types: Sine, Square, Triangle, Saw and Circle. More can be added implementing " + PeriodicSignal.class.getName() 
			+ " interface. If output type is chosen file or serial then name must be specified like COM1 for ports. Giving 'ser' wo name shows all known ports in the system. Data is sent at port defalult rate."
			+ " The defaults are: PCM: signed stereo 16bit. Signal: 0.7*sin(2Pi*1000) infinitely sampled at 44.1kHz and rendered on audio interface.");
		System.exit(1);
	}


	private static void log(String strMessage) {
		System.out.println(strMessage);
	}
}

