/**

	Valentin Tihomirov, 2006, www.abelectron.com

	
Changes to OscillatorPlayer:
 The table were replaced by continous sampling. Output signal quality is much better.

TODO: 
 - options for samplebits and endians
 - weired thing is that sound is much poorer when playing as low as
 	22050 frames and improfes quantumly when playing one frame more.
*/

import java.io.IOException;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.SourceDataLine;

import gnu.getopt.Getopt;


class Console {

	public static void main(String[] args) throws Exception {
		
		final int	BUFFER_SIZE = 128000;
		byte nChannels = 2; // actually we need mono but on one of my audio cards wierdly bubbles in the mono mode
		
		PeriodicSignal oscillator = new Sine(); // play sin by default
		float	sampleRate = 44100.0F;
		float	signalFrequency = 1000.0F;
		float	amplitude = 0.7F;
		int		sampleSizeInBits = 16; // 8 or 16
		float duration = 0; // infinity
		
		//
		// Parse arguments
		//
		
		if (args.length == 0)
			log("use -? option for usage");
		else {
			Getopt opt = new Getopt(Console.class.getName(), args, "?mo:r:f:a:d:");
			int	c;
			while ((c = opt.getopt()) != -1) {
				switch (c) {
					case '?': printUsageAndExit();
					case 'm': nChannels = 1; break;
					case 'o': oscillator = (PeriodicSignal) Class.forName(opt.getOptarg()).newInstance(); break;
					case 'r': sampleRate = Float.parseFloat(opt.getOptarg()); break;
					case 'f': signalFrequency = Float.parseFloat(opt.getOptarg()); break;
					case 'a': amplitude = Float.parseFloat(opt.getOptarg()); break;
					//case 'b': sampleSizeInBits = Integer.parseInt(opt.getOptarg()); break;
					case 'd': duration = Float.parseFloat(opt.getOptarg()); break;
					default: break;
				}
			}
		}
		
		long remainingFrames = remainingFrames = (duration == 0) ? AudioSystem.NOT_SPECIFIED : (long) (duration * sampleRate);
		//log(remainingFrames + " frames reamins");
		
		// 
		// Get Audio Line
		//
		
		final int frameSize = sampleSizeInBits/8 * nChannels;
		final AudioFormat audioFormat = new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, sampleRate, sampleSizeInBits, nChannels, frameSize, sampleRate, false);
		DataLine.Info info = new DataLine.Info(SourceDataLine.class, audioFormat);
		SourceDataLine line = (SourceDataLine) AudioSystem.getLine(info);
		line.open(audioFormat);
		try {
			
			amplitude = (float) (amplitude * Math.pow(2, sampleSizeInBits-1) -1);
			final float dt = signalFrequency/sampleRate;
			float t = 0;
			byte[] buf = new byte[BUFFER_SIZE];
			
			line.start();
		
			log ("press Contorl-C to terminate");
			
			while_loop:
			while (true) {
				
				// signal sampling for-loop
				for (int i = 0; i < buf.length; i += frameSize) {
					
					if (remainingFrames != AudioSystem.NOT_SPECIFIED && remainingFrames-- == 0) {
						//log("flushing " + i/frameSize + " frames");
						//remainingFrames == 0;
						line.write(buf, 0, i); // play the tail and exit
						break while_loop;
					}
					
					short sample = (short) Math.round(oscillator.getValue(t) * amplitude);
			
					// this is for 16 bit little endian
					buf[i + 0] =  (byte) (sample & 0xFF);
					buf[i + 1] = (byte) ((sample >>> 8) & 0xFF);
		
					if (nChannels == 2)	{
						buf[i + 2] = buf[i + 0];
						buf[i + 3] = buf[i + 1];
					}
					
					if ((t += dt) > 1) // (t < 1) + (dt < 1) < 2
						t -= 1;
				}
				
				// audio card output
				line.write(buf, 0, buf.length);
			}
				
		} finally {
			line.close();
		}
		
	}



	static void printUsageAndExit() {
		log("usage: java " + Console.class.getName() + " [-m] [-o <waveformtype>] [-a <amplitude>] [-f <signalfrequency>] [-r <samplerate>] [-d <duration>]");
		log("\tamplitude: 0..1");
		log("\tm = mono");
		log("\toscillator types: Sine, Square, Triangle, Saw and Circle. More can be added implementing " + PeriodicSignal.class.getName() + " interface.");
		System.exit(1);
	}


	private static void log(String strMessage) {
		System.out.println(strMessage);
	}
}

