import javax.sound.sampled.AudioFormat;


class Player {
	
	public static void log(String msg) {
		Gui.log(msg);
	}
	
	// these vars have a different (not-sync) declaration in alternative Player,
	// therefore, we cannot share them in Gui. But the names are the same, so
	// we can let the Gui to set them up.
	static volatile PeriodicSignal signal;
	static volatile float amplitude;
	static volatile float signalFrequency;
	
	//Gui event handlers
	public static void windowClosed() {	done = true; }
	public static void setDutyCycle(float dc) {
		// avoid collistions with the main thread samplig the signal
		synchronized (Player.signalLock) {((DutyCycle) signal).setDutyCycle(dc);}
	}
	public static void setSignal(PeriodicSignal signal)  {
		synchronized (signalLock) {Player.signal = signal;}
	}
	public static void onAmplitudeChange() {}
	public static void onFrequencyChange() {}
	
	// additional options exposed by Gui
	public static String options = "-Dre=<reaction time>";
	public static String optionDescription = "Reaction time is a floating point number of seconds of sweeping used to smooth the parameter adjustmemts.";
	
	
	private static volatile boolean done = false;
	private final static Object signalLock = new Object();

		
	// the playback loop called by main thread once gui form is shown
	public static void pump() throws java.io.IOException {
	/*Should the buf and line be Gui members?*/
	
	
		float responseTime = 1; // time in seconds. Must translate from 0 to max (or 1 to max in log scale) in this time.
		
		{ // extract properties
			String prop = System.getProperty("re");
			if (prop != null)
				if (Float.parseFloat(prop) < 0)
					log("Sample response time cannot be negative");
				else
					responseTime = Float.parseFloat(prop);
			//else log("Using default reaction time = " + responseTime + " sec");

		}
		
			
		//
		// Pump audio data from input stream to output line. Gui will cease the input stream on exit.
		//
		
		float t = 0;
		
		
// 			// merely uses immidiate signalFrequency and amplitude
// 		while (!done) {
// 			
// 			// sample
// 			final float a = (float) (amplitude * Math.pow(2, sampleSizeInBits-1) -1);
// 			
// 			
// 			final float dt = signalFrequency/sampleRate;
// 			
// 			synchronized(signalLock) {
// 				for (int i = 0; i < buf.length; i += frameSize) {
// 					short sample = (short) Math.round(a * signal.getValue(t));
// 			
// 					buf[i + 0] =  (byte) (sample & 0xFF);
// 					buf[i + 1] = ...
// 					
// 					if ((t += dt) > 1) // (t < 1) + (dt < 1) < 2
// 						t -= 1;
// 				}
// 			}
// 			line.write(buf, 0, buf.length);
// 		}

			
						
		// smooth freq transition. The options are exp(y) response, oscillating at w0 resp 
		// (degenerating to pure exp when y<w0), circle, Sigma (cos, for example), line, 
		// poly trend interpolation. Linear approximation in log scale was chosen.
		float actualFrequency = signalFrequency; // smooth frequency transition from actualFrequency to desired signalFrequency
		final double df = Math.pow(Gui.maxFreq, 1/(responseTime * Gui.line.getFormat().getSampleRate()+1)); // (minfreq = 1) * x^cycles = MaxFreq, hence x = maxf^(1/cycles)
		
		// amplitude scale is linear.
		float actualAmp = amplitude; // smooth amp trans
		final float da = 1/(responseTime * Gui.line.getFormat().getSampleRate()+1); // amax - amin = 1-0 = 1;
		final double maxA = Math.pow(2, Gui.line.getFormat().getSampleSizeInBits()-1);
		
						
		
		//while (line.isActive()) {
		while (!done) {
			
			// sample
			synchronized(signalLock) {
				for (int i = 0; i < Gui.buf.length; i += Gui.line.getFormat().getFrameSize()) {
					
					// adjust amplitude
					if ((actualAmp < amplitude && (actualAmp += da) >  amplitude) 
					 || (actualAmp > amplitude && (actualAmp -= da) <  amplitude))
						actualAmp = amplitude;

												
					final float a = (float) (actualAmp * maxA - 1);
					short sample = (short) Math.round(a * signal.getValue(t));
			
					// this is for 16 bit little endian
					for (int channel = 0; channel < Gui.line.getFormat().getChannels(); channel++) {
						Gui.buf[i++] = (byte) (sample & 0xFF);
						if (Gui.line.getFormat().getSampleSizeInBits() == 16)
							Gui.buf[i++] = (byte) ((sample >>> 8) & 0xFF);
					}
/*					Gui.buf[i + 0] =  (byte) (sample & 0xFF);
					Gui.buf[i + 1] = (byte) ((sample >>> 8) & 0xFF);
		
					if (Gui.line.getFormat().getChannels() == 2)	{
						Gui.buf[i + 2] = Gui.buf[i + 0];
						Gui.buf[i + 3] = Gui.buf[i + 1];
					}*/
					
					
					// adjust freq
					if ((actualFrequency < signalFrequency && (actualFrequency *= df) >  signalFrequency) 
					 || (actualFrequency > signalFrequency && (actualFrequency /= df) <  signalFrequency))
						actualFrequency = signalFrequency;
					
						
					// advance time	to next sample
					float dt = actualFrequency/Gui.line.getFormat().getSampleRate(); 
					if ((t += dt) > 1) // (t < 1) + (dt < 1) < 2
						t -= 1;
				}
			}
			
			//Gui.line.write(buf, 0, buf.length);
			Gui.bufToLine(Gui.buf.length);
		}
	}
}