// Valentin Tihomirov, 2006, www.abelectron.com

/*

This Gui is shared among two Players: the first plays back a table of a period of samples of 
	a signal. Another, the fluent, does not relay on a table and samples the signal as many
	times as needed allowing for sound-modulation effects in the process.
	
	The Gui-Player are class-path statically bound. That is, you have to put both
	on the classpath and execute theGui main.

	// If it is hard to grip for you what is going on here, consider the non-gui version first.

*/


/*ToDo: 
	- we can lock on the mandatory Sine signal instead of special object.
	- signal frequency and amplitude may be done protected by the same signal lock. Is it better than volatile?
	*/


// import titlenter.*;
// import floatmodel.*;

import java.io.IOException;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.SourceDataLine;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.GeneralPath;
import javax.swing.*;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;

import java.util.*;
import java.io.*;

/*
This shared gui has the following changes from floating gui version:
	- all procs and vars become non-static, the main(args) renamed to constructor
	- signal, amplitude, frequency and done made public.
	- 
*/

/**In general, I wanted to make just gui changing signal, amplitude and frequency, which 
would be instantiated by the main program. However, I have realized that there are other 
code parts (like buffer size and other audio line parameters) which should be the same
among graphic programs. So just put the Oscillator on the path and the gui will instantiate 
it as Strategy pattern.
*/
public class Gui {

		
	static JPanel label(String name, JSlider slider) {
		BorderLayout bl = new BorderLayout();
		bl.setHgap(10);
		JPanel panel = new JPanel(bl);
		panel.add(new JLabel(name), BorderLayout.LINE_START);
		panel.add(slider, BorderLayout.CENTER);
		return panel;
	}

	public static float maxFreq;
	public static SourceDataLine line;
	public static byte[] buf;
	private static OutputStream file;
	
	public static void main(String[] args) throws Exception {
		
		// default params
		Player.signal = new Sine();
		Player.amplitude = 0.7F;
		Player.signalFrequency = 440F;//2949.2769f;//3041.192F; //440.0F;
		
		int bufferSize = 12800; //128000;
		byte nChannels = 1; // actually we need mono but on one of my audio cards the mono channel sound seems wierdly bubbling
		int sampleSizeInBits = 16; // 8 or 16
		float sampleRate = 44100.0F;
		
		
		
		
		///
		// parse cmdline arguments
		///
		
		for (int a = 0; a < args.length; a++) {
			if (args[a].compareTo("m") == 0)
				nChannels = 1;
			else if (args[a].compareTo("8") == 0)
				sampleSizeInBits = 8;
			else {
				log ("Usage: java -Dsr=<sampling rate> -Dbs=<buffer size> -Dof=<output file> " + Player.options + " " + Gui.class.getName() + " [8] [m]");
				log ("Default is 16-bit stereo, the optional '8' and 'm' change the format." + Utils.br() 
					+ "Sample rate is floating point number. " + Utils.br() 
					+ "Buffer size must be an integer divisible by 4." + Utils.br() 
					+ "Output file allows to record the sound produced in raw 8-bit mono." + Utils.br() 
					+ Player.optionDescription);
				System.exit(0);
			}
			
		}
		
		//log("Valentin Tihomirov, 2006, www.abelectron.com");
		//log("Type 'java " + Gui.class.getName() + " ?' for command line parameters");
		//log("To enter exact value, focus on slider and type digits");
		
		String prop;
		if ((prop = System.getProperty("sr")) == null)
			log ("Using default sample rate = " + sampleRate);
		else {
			sampleRate = Float.parseFloat(prop);
			if (sampleRate < Player.signalFrequency) {
				sampleRate = Player.signalFrequency;
				log("Sample rate is limited to " + Player.signalFrequency);
			}
		}
		
		maxFreq = sampleRate/3;
		
		if ((prop = System.getProperty("bs")) == null)
			log ("Using default buffer size = " + bufferSize);
		else
			bufferSize = Integer.parseInt(prop);
		
		///
		// Start the playback by obtaining output line, opening it and feeding with data from input stream.
		// The playback will be stopped and program exits when the input stream runs out (gui stops it on exit).
		///
		
		final int frameSize = sampleSizeInBits/8 * nChannels;
		final AudioFormat audioFormat = new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, sampleRate, sampleSizeInBits, nChannels, frameSize, sampleRate, false);
		line = (SourceDataLine) AudioSystem.getLine(new DataLine.Info(SourceDataLine.class, audioFormat));
		line.open(audioFormat, bufferSize);
		try {

			if (bufferSize != line.getBufferSize())
				log("Line buffer size = " + line.getBufferSize());
			
			
			
			
			
			
			
			////////////////////////////////////////////////////////////
			// Create the GUI dialog for controlling the input stream //
			////////////////////////////////////////////////////////////
			
			{
				//final JDialog dialog = new JDialog();
				final FloatTitleEditFrame dialog = new FloatTitleEditFrame(/*"Audio Signal Wave Generator"*/);
				final String defaultTitle = "Audio Signal Wave Generator";
				
				dialog.setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);		
				dialog.addWindowListener(new WindowAdapter() {
					public void windowClosed(WindowEvent e) {
						Player.windowClosed();
					}
				});
				Container controlsPanel = new JPanel();
				dialog.getContentPane().add(controlsPanel, BorderLayout.PAGE_START);
				controlsPanel.setLayout(new BoxLayout(controlsPanel, BoxLayout.PAGE_AXIS)); // the weired fucking construction 
				
				
				class Plot extends ImageComponent2 {
		
					//Painting only the clipped region requires the signals to be time-independent (stateless). Tyristors are not. Their state can be 
					// opened or closed depending on the history. The requirement can be eliminated though if the canvas once rendered could redraw it 
					// from a buffer, since user always updates the plot as a whole and system never modifies the waveform! Lets draw an image every 
					// time we manipulate the signal by the gui controls. ImageComponent does the trick.
					
					PeriodicSignal signal = new Sine(); // a copy for gui drawing not to infere with audio output in main thread

					public void render(Graphics g) {
						
						
						//synchronized (signalLock) { // it is better to use a copy of signal instead. Causes serious sound lags, especially on large window.
							
							// the graphics area = [0,0,w-1,h-1]
							final int w = getWidth();
							final int h = getHeight();
							
							// translates between screen coordinates (0,0; w-1,h-1) and the signal definition area (0,-1; 1,1).
					        class Scaler {
						        // linearly scales screen x downto period 0..1, obtains signal value in the range -1..1, applies amplitude and converts it back to screen y coordinate so that (-1 => h-1) goes to bottom of the screen and (1 => 0) mapped to top of the screen)
						        int sample(int x) { return (int) ((-Player.amplitude*signal.getValue(x/(float)(w-1))+1)/2*(h-1));}
					        }
					        Scaler s = new Scaler();
					        
					        //g.setColor(Color.green);
					        //((Graphics2D) g).setStroke(new BasicStroke(3));
					        GeneralPath gp = new GeneralPath();
					        gp.moveTo(0, s.sample(0));
					        for (int x = 1; x < w; x++) //iterate the screen from left to right
						        gp.lineTo(x, s.sample(x));
					        ((Graphics2D) g).draw(gp);
			        	//}
					}
					
					public Plot() {	setPreferredSize(new Dimension(getWidth(), 100)); }
					
				}
				final Plot plot = new Plot();
				
		
						
// 				// not editable slider
// 				final JSlider dutySlider = new JSlider();
// 				dutySlider.setValue((int) (((DutyCycle) Player.signal).getDutyCycle()*100));
// 				dutySlider.addFocusListener(new FocusAdapter() {
// 						public void focusGained(FocusEvent e) {
// 							dialog.setTitle("duty cycle = " + dutySlider.getValue()/100f);
// 						}
// 				});
// 				final ChangeListener dutySliderChangeListener = new ChangeListener() {
// 					public void stateChanged(ChangeEvent e) {
// 						float dc = dutySlider.getValue()/100f; // scale -100..100 downto -1..1
// 						dialog.setTitle("duty cycle = " + dc);
// 						Player.setDutyCycle(dc);
// 						((DutyCycle) plot.signal).setDutyCycle(dc); //NB! this is not Player.signal
// 						plot.repaint();
// 		    		}
// 				};
// 				dutySlider.addChangeListener(dutySliderChangeListener);
// 				dutySlider.setEnabled(Player.signal instanceof DutyCycle);


				final IntFloatBoundedRangeAdapter dutyModel = new IntFloatBoundedRangeAdapter(
					new DefaultFloatBoundedRangeModel(((DutyCycle) Player.signal).getDutyCycle()*100, 0, 100)
					, 101);
				final FloatTitleEditFrame.Slider dutySlider = dialog.new DefaultFormatSlider(dutyModel, "Duty Cycle", "%");
				final ChangeListener dutySliderChangeListener = new ChangeListener() {
					public void stateChanged(ChangeEvent e) {
						float dc = dutySlider.getFloatValue()/100f; // scale -100..100 downto -1..1
						Player.setDutyCycle(dc);
						((DutyCycle) plot.signal).setDutyCycle(dc); //NB! this is not Player.signal
						plot.repaint();
		    		}
				};
				dutySlider.addChangeListener(dutySliderChangeListener);
				dutySlider.setEnabled(Player.signal instanceof DutyCycle);



				
				// generate wavoform buttons by scanning the directory for signal plugins
				final ButtonGroup signalButtonGroup = new ButtonGroup();
			    final JPanel signalButtonsPanel = new JPanel();
			    
				
				class SignalButton extends JRadioButton {
					final PeriodicSignal signal;
					final PeriodicSignal canvasSignal; // a copy for gui
					SignalButton(Class signalClass) throws Exception {
						super(signalClass.getName()); 
						signal = (PeriodicSignal) signalClass.newInstance();
						canvasSignal = (PeriodicSignal) signalClass.newInstance();
						signalButtonGroup.add(this);
						//addActionListener(signalButtonListener);
						if (signal.getClass() == Player.signal.getClass())
							setSelected(true);
						addActionListener(new ActionListener() {
							
							public void actionPerformed(ActionEvent e) {
								plot.signal = canvasSignal;
								plot.repaint();
								boolean b = signal instanceof DutyCycle;
								dutySlider.setEnabled(b);
								Player.setSignal(signal);
								if (b) {
									dutySlider.removeChangeListener(dutySliderChangeListener); // do we need to supress this event propagation?
									dutySlider.setFloatValue(((DutyCycle) signal).getDutyCycle()*100);
									dutySlider.addChangeListener(dutySliderChangeListener);
								}
								dialog.setTitle(defaultTitle);
							}
						});
						signalButtonsPanel.add(this);
					}
				}
		
				java.net.URL url = ClassLoader.getSystemResource(Player.signal.getClass().getName() + ".class");
				File signalsDir = new File(new java.net.URI(url.toExternalForm())).getParentFile();
				//File signalsDir = new File(url.getFile()).getParentFile();
				
				//(new File(".")).listFiles(new FileFilter() { 
				signalsDir.listFiles(new FileFilter() { 
					public boolean accept(File file) {
						
						if (file.isFile() && file.getName().endsWith(".class")) // if it looks like java class
							try { 												
								final Class c =	Class.forName(Utils.getBase(file));	// try loading the class 
								if (PeriodicSignal.class.isAssignableFrom(c) && Utils.isInstantiable(c)) // if the class looks like our plug-in 
									new SignalButton(c); // try instantiating altogether with its selecting button
							} catch (Throwable e) {
								log(e.getClass().getName() + " plugging signal '" + file.getName() + "' in with a msg: '" + e.getMessage() + "'");
							}

							
						return false;
					}
				});				
				
				final IntFloatBoundedRangeAdapter amplModel = new IntFloatBoundedRangeAdapter(
					new DefaultFloatBoundedRangeModel(Player.amplitude*100, 0, 100)
					, 101);
				final JSlider amplitudeSlider = dialog.new DefaultFormatSlider(amplModel, "Amplitude", "%");
				amplitudeSlider.addChangeListener(new ChangeListener() {
					public void stateChanged(ChangeEvent e) {
						Player.amplitude = amplModel.getFloatValue()/100f; // scale -100..100 downto -1..1
						Player.onAmplitudeChange();
						plot.repaint();
		    		}
				});
		
				
 				final int GRADES = 1000;
 				//final FloatBoundedRangeModel logModel = new LogBoundedRangeModel((float) Math.pow(maxFreq, 1/(double)GRADES), 1f, 0f, 0, GRADES, 0f, (float) maxFreq);
				final IntFloatBoundedRangeAdapter logModel = new IntFloatLogAdapter(
					new DefaultFloatBoundedRangeModel(440, 0, maxFreq)
					, GRADES, (float) Math.pow(maxFreq, 1/(double)(GRADES-1)), 1f, 0f);
 				final JSlider frequencySlider = dialog.new DefaultFormatSlider(logModel, "Frequency", "Hz");
 				logModel.setFloatValue(Player.signalFrequency);
 				logModel.addChangeListener(new ChangeListener() {
 					public void stateChanged(ChangeEvent e) {
 						Player.signalFrequency = logModel.getFloatValue();
						Player.onFrequencyChange();
 		    		}
 				});
 				
				
				
				controlsPanel.add(signalButtonsPanel);
				controlsPanel.add(label("amplitude", amplitudeSlider));
				controlsPanel.add(label("duty cycle", dutySlider));
				controlsPanel.add(label("frequency", frequencySlider));
			
				dialog.getContentPane().add(plot, BorderLayout.CENTER);
				
				dialog.pack();
				dialog.setTitle(defaultTitle);
				dialog.setVisible(true);
				// no more gui in this thread
			}	
			
			
			
			Thread.currentThread().setPriority(Thread.MAX_PRIORITY);
			
			line.start();
			
			{// extract properties
				if ((prop = System.getProperty("of")) != null) {
					
					log("Recording into '" + prop + "'"); //String fname = "8bit.mono";
					
					class Pass8BitMonoFilter extends FilterOutputStream {
						Pass8BitMonoFilter(OutputStream out) {super(out);}
						public void write(int b) throws IOException {throw new RuntimeException("method not implemented");}
						public void write(byte[] b, int off, int len) throws IOException{
							if (len % Gui.line.getFormat().getFrameSize() != 0) 
								throw new IOException("length must be an integer multiple of frame size");
							int stop = off + len;
							for (int i = off + 1; i < stop; ) {
								super.write(b[i]);
								i += Gui.line.getFormat().getFrameSize();
							}
						}
					}
					
					file = new Pass8BitMonoFilter(new BufferedOutputStream(new FileOutputStream(prop)));
					
				} else {
					
					file = new OutputStream() { //NullOutputStream();
						public void write(int b) {throw new RuntimeException("method not implemented");}
						public void write(byte[] b, int off, int len) {}
					};
				}
			}
				
				
			try { // finally close file
				buf = new byte[line.getBufferSize()];
				Player.pump();
			} finally {
				file.close();
			}
				
		} finally {
			line.close();
		}
		
		

	}

	public static void bufToLine(int len) throws IOException {
		file.write(buf, 0, len);
		Gui.line.write(buf, 0, len);
	}
	
	public static void log(String strMessage) {
		System.out.println(strMessage);
	}
}





/**Fun*/
/*class Gui2 extends Thread {
	
	public void run() {
		try {
			Gui.main(new String[0]);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public static void main(String[] args) throws Exception {
		Thread t1 = new Gui2();
		Thread t2 = new Gui2();
		t1.start();
		t2.start();
		t1.join();
		t2.join();
	}
}*/




