import java.io.*;
import java.util.*;

/**/
class Mixer implements PeriodicSignal {
	
	class Bean {
		final PeriodicSignal waveform;
		float aMult;
		final int fMult;
		Bean(PeriodicSignal waveform, float aMult, int fMult) {this.waveform = waveform; this.aMult = aMult; this.fMult = fMult;}
	}
	
	Collection<Bean> mixture = new ArrayList<Bean>();
	float maxAmpl = 0;

	/**THe constructor parses the config file*/
	Mixer() throws IOException {
		//BufferedReader br = new BufferedReader(new FileReader("mixer.cfg"));
		BufferedReader br = new BufferedReader(new InputStreamReader(this.getClass().getClassLoader().getResourceAsStream( "mixer.cfg" )));
		
		try {
			for (int lineN = 1; ; lineN++) {
				
				String line = br.readLine();
				if (line == null)
					break;
					
				line = line.trim();
				
				if (line.length() == 0 || line.startsWith("#"))
					continue;
				
				try {
					String[] strBean = line.split("\\s+");
					int b;
					PeriodicSignal wf;
					if (strBean.length == 2) {
						wf = new Sine();
						b = 0;
					} else {
						wf = (PeriodicSignal) Class.forName(strBean[0]).newInstance();
						b = 1;
					}
					float ampl = Float.parseFloat(strBean[b]);
					mixture.add(new Bean(
						wf,
						ampl,
						Integer.parseInt(strBean[b+1])
					));
					maxAmpl += ampl;
				} catch (Exception e) {
					System.out.println("Could not parse line " + lineN + ": '" + line + "'");
				}
			}
			
			// normalize amplitudes
			for (Bean bean : mixture)
				bean.aMult = bean.aMult/maxAmpl;
			
			//log("maxAmpl = " + maxAmpl);
			
		} finally {
			br.close(); // what if an exception here?
		}
		
	}
	
	/**Scalar multiplication of weight vectors by corresponding function values at phase t*/
	public float getValue(float t) {
		float result = 0;
		for (Bean bean : mixture)
			result += bean.aMult*bean.waveform.getValue((bean.fMult * t) % 1);
		return result;
	};
	
	private static void log(String msg) {
		System.out.println(msg);
	}
}