import javax.sound.sampled.*;

/**Incapsulates audioformat*/
class AudioBuffer {
	
	public final byte[] table;
	public final AudioFormat format;
	
	AudioBuffer (AudioFormat format, byte[] table) {
		this.format = format;
		this.table = table;
	}
	
	static boolean stereo = true; // actually we need mono but on one of my audio cards the mono channel sound seems wierdly bubbling
	
	static AudioBuffer sample(PeriodicSignal signal, float amplitude, float signalFrequency, float sampleRate) {
		
		final int bps = 16; // 8 or 16
		final int nChannels = stereo ? 2 : 1;
		
		final AudioFormat audioFormat = new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, sampleRate, bps, nChannels, bps/8 * nChannels, sampleRate, false);
			
		// prepare waveform table by sampling a signal for 1 period
		
		// mapping reals [-1..1] to discretes [-32767..32767] as given in ava Sound Resources: FAQ: Audio Programming at http://www.jsresources.org/faq_audio.html#no_volume_control
		// 2^15 - 1 = 32767
		amplitude = (float) (amplitude * Math.pow(2, audioFormat.getSampleSizeInBits()-1) -1);
		final int framesPerPeriod = Math.round(audioFormat.getFrameRate() / signalFrequency); // length of one period in frames
		int tableLength = framesPerPeriod * audioFormat.getFrameSize();
		final byte[] table = new byte[tableLength]; // signal waveform table
		for (int frame = 0; frame < framesPerPeriod; frame++) {
			
			float periodPosition = (float) frame / (float) framesPerPeriod;
			short sample = (short) Math.round(signal.getValue(periodPosition) * amplitude);
			
			int baseAddr = frame * audioFormat.getFrameSize();
			
			
			// this is for 16 bit little endian
			table[baseAddr + 0] =  (byte) (sample & 0xFF);
			table[baseAddr + 1] = (byte) ((sample >>> 8) & 0xFF);

			if (stereo)	{
				table[baseAddr + 2] = table[baseAddr + 0];
				table[baseAddr + 3] = table[baseAddr + 1];
			}
			

		}
		
		return new AudioBuffer(audioFormat, table);
	}
}