/**
This is an optimized (clipped) version of player where AudioInputSpream was eliminated by switching output line from SourceDataLine to looping Clip.

Weired:
 - neither clip.start(0 nor clip.drain() seem to function
 - What is the difference between gain and volume, balance and pan?
 - Both balance and pan controls affect the loudness of both speakers synchronously increasing/decresing both. That is, they behave like volume control rather than migrate the sound between speakers.
 - Despite the master gain is claimed to limit the volume rather than wrapping around in the javadoc, it wraps the signal around.
 */

import java.io.IOException;

import javax.sound.sampled.Clip;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;



public class ConsoleLooper {

	public static void main(String[] args) throws Exception {
		
		ConsolePlayer.parseArguments(args);

		// sample the signal
		AudioBuffer audioBuffer = AudioBuffer.sample(
			(PeriodicSignal) ConsolePlayer.oscillatorClass.newInstance(), 
			ConsolePlayer.amplitude, 
			ConsolePlayer.signalFrequency, 
			ConsolePlayer.sampleRate);

		
				
		// create, open and start the infinite clip of the buffer:
		
		Clip line = (Clip) AudioSystem.getLine(new DataLine.Info(Clip.class, audioBuffer.format));
		line.open(audioBuffer.format, audioBuffer.table, 0, audioBuffer.table.length);
		try {
			
			//line.start(); // does not work
			//line.drain(); // does not work
			line.loop(Clip.LOOP_CONTINUOUSLY);

			log ("press Contorl-C to terminate");
			while (true) // stall
				Thread.sleep(5000);
			
		} finally {
			line.close();
		}
		
	}

	private static void log(String strMessage) {
		System.out.println(strMessage);
	}
}

