/**
Changes:
 - OOP style allowing for plug-in oscillator exptesibility;
 - signal functions re-emplemented;
 - the bug of PCM overflow on the max (1) signal amplitude.
 - circle and mixer signals was added. The mixer allows to compose new signals as series of basic ones (like Fourier sine composition) without compilation.
 

TODO: 
 -D duration (in seconds, remainingFrames = durationSeconds/frameRate)
 
Weired:
 - the same sine wave in stereo mode sounds clearly while some noise interferes in mono but everything is clear at oscillograph???
 - AudioSystem.getSourceLineInfo(), which is logical without parameters and presented at http://www.jsresources.org/faq_audio.html#mixers cannot be found by java 1.5 compiler!

 
*/

import java.io.IOException;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.SourceDataLine;

import gnu.getopt.Getopt;



public class ConsolePlayer {
	static final int	BUFFER_SIZE = 128000;
	static Class oscillatorClass = Sine.class; // play sin by default
	static float	sampleRate = 44100.0F;
	static float	signalFrequency = 1000.0F;
	static float	amplitude = 0.7F;

	static void parseArguments(String[] args) throws ClassNotFoundException {
		
		if (args.length == 0) {
			log("use -? option for usage");
			return;
		}
			
		Getopt opt = new Getopt(ConsolePlayer.class.getName(), args, "?mo:r:f:a:D");
		int	c;
		while ((c = opt.getopt()) != -1) {
			switch (c) {
				case '?': printUsageAndExit();
				case 'm': AudioBuffer.stereo = false; break;
				case 'o': oscillatorClass = Class.forName(opt.getOptarg()); break;
				case 'r': sampleRate = Float.parseFloat(opt.getOptarg()); break;
				case 'f': signalFrequency = Float.parseFloat(opt.getOptarg()); break;
				case 'a': amplitude = Float.parseFloat(opt.getOptarg()); break;
				case 'D': break;
				default: break;
			}
		}
	}
			
	public static void main(String[] args) throws Exception {
		
		parseArguments(args);
		
		// prepare input buffer
		AudioInputStream in = new OscillatorInputStream(
			AudioBuffer.sample(
				(PeriodicSignal) oscillatorClass.newInstance(),
				amplitude,
				signalFrequency,
				sampleRate),
			AudioSystem.NOT_SPECIFIED);
			
		// obtain output line, open and write to it
		AudioFormat audioFormat = in.getFormat();
		//audioFormat = new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, audioFormat.getSampleRate(), audioFormat.getSampleSizeInBits(), audioFormat.getChannels(), audioFormat.getFrameSize(), audioFormat.getFrameRate(), audioFormat.isBigEndian());
		DataLine.Info info = new DataLine.Info(SourceDataLine.class, audioFormat);
		SourceDataLine line = (SourceDataLine) AudioSystem.getLine(info);
		line.open(audioFormat);
		try {
			
			line.start();
		
			log ("press Contorl-C to terminate");
			byte[] buf = new byte[BUFFER_SIZE];
			int	numBytesRead;
			while ((numBytesRead = in.read(buf)) != -1)
				line.write(buf, 0, numBytesRead);
				
		} finally {
			line.close();
		}
		
	}



	static void printUsageAndExit() {
		log("usage:");
		log("\tjava " + ConsolePlayer.class.getName() + " [-o <waveformtype>] [-a <amplitude>] [-f <signalfrequency>] [-r <samplerate>] [-m]");
		log("\tamplitude: 0..1");
		log("\tm for mono, stereo is default since mono is buggy on my audiocard.");
		log("\toscillator types: Sine, Square, Triangle, Saw and Circle. More can be added implementing " + PeriodicSignal.class.getName() + " interface.");
		System.exit(1);
	}


	private static void log(String strMessage) {
		System.out.println(strMessage);
	}
}

