
import java.io.ByteArrayInputStream;
import java.io.IOException;

import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioFormat;

public class OscillatorInputStream	extends AudioInputStream {

	private byte[] table;
	private int tablePosition = 0;
	public volatile long remainingFrames; // accessed concurrently by reader and stopper
	//long remainingFrames;
	
	synchronized void setWaveformTable(byte[] table) {
		//tablePosition = 0;
		if (this.table != null) {
			framePos = tablePosition/frameSize;	// get pos in old table
			framePos = framePos * table.length / this.table.length; // jump to the same offset in new period
			tablePosition = (int) framePos * frameSize;
		}
		
		this.table = table;
		
	}

	public OscillatorInputStream(AudioFormat audioFormat, byte[] table, long remainingFrames) {
		
		super(new ByteArrayInputStream(new byte[0]), audioFormat, remainingFrames);
		this.remainingFrames = remainingFrames;
		setWaveformTable(table);
	}


	/**	Returns the number of bytes that can be read without blocking.
		Since there is no blocking possible here, we simply try to
		return the number of bytes available at all. In case the
		length of the stream is indefinite, we return the highest
		number that can be represented in an integer. If the length
		if finite, this length is returned, clipped by the maximum
		that can be represented.
	*/
	public int available() {
		return remainingFrames == AudioSystem.NOT_SPECIFIED ? Integer.MAX_VALUE	: (int) Math.min(remainingFrames * getFormat().getFrameSize(), (long) Integer.MAX_VALUE);
	}



	/*
	  this method should throw an IOException if the frame size is not 1.
	  Since we currently always use 16 bit samples, the frame size is
	  always greater than 1. So we always throw an exception.
	*/
	public int read() throws IOException {
		throw new IOException("cannot use this method currently");
	}



	public synchronized int read(byte[] buf, int off, int len) throws IOException {
		
		if (len % frameSize != 0)//getFormat().getFrameSize() != 0)
			throw new IOException("length must be an integer multiple of frame size");
			
		int	constrainedLength = Math.min(available(), len);
		int	remainingLength = constrainedLength;
		while (remainingLength > 0) {
			int	numBytesToCopyNow = Math.min(table.length - tablePosition, remainingLength);
			System.arraycopy(table, tablePosition, buf, off, numBytesToCopyNow);
			remainingLength -= numBytesToCopyNow;
			off += numBytesToCopyNow;
			tablePosition = (tablePosition + numBytesToCopyNow) % table.length;
		}
		int	framesRead = constrainedLength / frameSize;//getFormat().getFrameSize();
		
		if (remainingFrames != AudioSystem.NOT_SPECIFIED)
			remainingFrames -= framesRead;
			
		return remainingFrames == 0 ? -1 : constrainedLength;
		
	}


	static void log(String strMessage) {
		System.out.println(strMessage);
	}
}





