/**
experimantal version
*/

import java.io.IOException;

import javax.sound.sampled.*;
import javax.swing.*;
import javax.swing.event.*;
import java.awt.event.*;
import java.util.*;
import java.awt.BorderLayout;
import java.io.*;





public class OscillatorLooper {
	private static final int	BUFFER_SIZE = 1280; // was 128000
	
	static byte[] table;
	static AudioFormat audioFormat;
	static boolean stereo = true; // actually we need mono but on one of my audio cards the mono channel sound seems wierdly bubbling
	static int framesPerPeriod;
	static float amplitude = 0.7F;
	static Clip line;
	static PeriodicSignal signal = new Sine();

	
	static void obtainAndOpenLine() throws Exception {
		line = (Clip) AudioSystem.getLine(new DataLine.Info(Clip.class, audioFormat));
		line.open(audioFormat, table, 0, table.length);
	}

	static void updateTable() {
		float a = (float) (amplitude * Math.pow(2, audioFormat.getSampleSizeInBits()-1) -1);
		for (int frame = 0; frame < framesPerPeriod; frame++) {
			
			float periodPosition = (float) frame / (float) framesPerPeriod;
			short sample = (short) Math.round(signal.getValue(periodPosition) * a);
			
			int baseAddr = frame * audioFormat.getFrameSize();
			
			// this is for 16 bit little endian
			table[baseAddr + 0] =  (byte) (sample & 0xFF);
			table[baseAddr + 1] = (byte) ((sample >>> 8) & 0xFF);
	
			if (stereo)	{
				table[baseAddr + 2] = table[baseAddr + 0];
				table[baseAddr + 3] = table[baseAddr + 1];
			}
		}
	}
	
	public static void main(String[] args) throws Exception {

		Class oscillatorClass = Sine.class;
		float	sampleRate = 44100.0F;
		float	signalFrequency = 1000.0F;

		// Line.Info[] infos = AudioSystem.getTargetLineInfo();
		//printLineInfos(AudioSystem.getSourceLineInfo(new Line.Info(SourceDataLine.class))); // as well as clip?
		//printLineInfos(AudioSystem.getTargetLineInfo(new Line.Info(TargetDataLine.class)));

		audioFormat = stereo ? 
				new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, sampleRate, 16, 2, 4, sampleRate, false)
			:
				new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, sampleRate, 16, 1, 2, sampleRate, false);
		
			
		// prepare waveform table by sampling a signal for 1 period
		
		// mapping reals [-1..1] to discretes [-32767..32767] as given in ava Sound Resources: FAQ: Audio Programming at http://www.jsresources.org/faq_audio.html#no_volume_control
		// 2^15 - 1 = 32767
		//amplitude = (float) (amplitude * Math.pow(2, audioFormat.getSampleSizeInBits()-1) -1);
		framesPerPeriod = Math.round(audioFormat.getFrameRate() / signalFrequency); // length of one period in frames
		int tableLength = framesPerPeriod * audioFormat.getFrameSize();
		table = new byte[tableLength]; // signal waveform table
		
		
		// create, open and start the infinite clip:
		obtainAndOpenLine();
		try {
			
			//line.start(); // does not work
			//line.drain(); // does not work
			line.loop(Clip.LOOP_CONTINUOUSLY);

						
			Control[] controls = line.getControls();
			log("obtained " + controls.length + " controls:");
			for (int i = 0; i < controls.length; i++)
				log("  [" + i + "] " + controls[i]);
				
			final FloatControl gain = (FloatControl) line.getControl(FloatControl.Type.MASTER_GAIN);
			log("gain = " + gain.getValue());
			
			
			//gain.shift(0, gain.getMaximum(),1000000); Thread.sleep(1000); log("gain = " + gain.getValue()); // this experiment shows that 2x gain wrapps around (even no limiting implemented)
			//gain.shift(gain.getMaximum(), gain.getMinimum(), 2000000); Thread.sleep(2000); log("gain = " + gain.getValue());
			//gain.shift(gain.getMinimum(), 0, 1000000); Thread.sleep(1000); log("gain = " + gain.getValue());
			
			
			
			// Create dialog
			
			final JDialog dialog = new JDialog();
			dialog.setTitle("Gain slider cannot be set exactly to current value " + gain.getValue() + gain.getUnits() + " cos it is discrete.");
			dialog.getContentPane().setLayout(new BoxLayout(dialog.getContentPane(), BoxLayout.PAGE_AXIS));

						
			// add gain slider
			
			//final JSlider slider = new JSlider();
			final JSlider gainSlider = new JSlider(0, 500, 450); // acutually, the gain is floating and max = 6.0206
			gainSlider.addChangeListener(new javax.swing.event.ChangeListener() {
				public void stateChanged(ChangeEvent e) { //if (!slider.getValueIsAdjusting()) {}
					float g = (gain.getMaximum() - gain.getMinimum())/(gainSlider.getMaximum() - gainSlider.getMinimum())*(gainSlider.getValue() - gainSlider.getMinimum()) + gain.getMinimum();
					dialog.setTitle(g + gain.getUnits());
	    			gain.setValue(g);
        		}
			});
			dialog.getContentPane().add(gainSlider/*, BorderLayout.CENTER*/);
			
			// add balance - pan sliders. What is the difference?
			if (stereo) {
				
				final JPanel pane = new JPanel();
				pane.setLayout(new BoxLayout(pane, BoxLayout.LINE_AXIS));
				
				class FloatControlSlider {
					void add(FloatControl.Type type) {
						final FloatControl control = (FloatControl) line.getControl(type);
						final JSlider slider = new JSlider(-100, 100, 0);
						slider.addChangeListener(new javax.swing.event.ChangeListener() {
							public void stateChanged(ChangeEvent e) { //if (!slider.getValueIsAdjusting()) {}
								float f = slider.getValue()/100f; // scale -100..100 downto -1..1
								dialog.setTitle(control + ", setting value " + f);
				    			control.setValue(f);
			        		}
						});
						pane.add(slider);
					}
				}
				FloatControlSlider helper = new FloatControlSlider();
				helper.add(FloatControl.Type.BALANCE);
				helper.add(FloatControl.Type.PAN);
				dialog.getContentPane().add(pane/*, BorderLayout.CENTER*/);
				
			}
			
			
			// add buttons
			
			final ButtonGroup signalButtonGroup = new ButtonGroup();
		    JPanel signalButtonsPanel = new JPanel();
		    ActionListener signalButtonListener = new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					try {
						signal = (PeriodicSignal) Class.forName(((JRadioButton) e.getSource()).getText()).newInstance();
						updateTable();
					} catch (Exception ex) {
						ex.printStackTrace();
					}
				}
			};

		    
		    JRadioButton sineButton = new JRadioButton("Sine");
		    signalButtonGroup.add(sineButton);
		    signalButtonGroup.add(new JRadioButton("Circle"));
		    signalButtonGroup.add(new JRadioButton("Square"));
		    signalButtonGroup.add(new JRadioButton("Triangle"));
		    signalButtonGroup.add(new JRadioButton("Saw"));
			for (Enumeration e = signalButtonGroup.getElements() ; e.hasMoreElements() ;) {
				JRadioButton b = (JRadioButton) e.nextElement();
				b.addActionListener(signalButtonListener);
				signalButtonsPanel.add(b);
			}
			sineButton.doClick(); // press to start mudic
			
			
			//amplitude, requires recalculation of buffer
			final JSlider amplitudeSlider = new JSlider(0, 100);
			amplitudeSlider.addChangeListener(new javax.swing.event.ChangeListener() {
				public void stateChanged(ChangeEvent e) { //if (!slider.getValueIsAdjusting()) {}
					amplitude = amplitudeSlider.getValue()/100f; // scale -100..100 downto -1..1
					dialog.setTitle("amplitude = " + amplitude);
					//Utils.getSelectedJRadioButton(signalButtonGroup).doClick();
					updateTable();
        		}
			});
			dialog.getContentPane().add(amplitudeSlider/*, BorderLayout.CENTER*/);
			
			
			dialog.getContentPane().add(signalButtonsPanel/*, BorderLayout.NORTH*/);
			
			
			/*
			
			// scan for classes on disk and create appropriate buttons
			
			File[] signalClassesFileList = (new File(".")).listFiles();
			log ("getting signal classes");
			for (int i = 0; i < signalClassesFileList.length; i++) {
				File f = signalClassesFileList[i];
				//log ("  [" + i + "] " + f.getName() + ", isFile:" + f.isFile());
				if (f.isFile() && Utils.getExtension(f).compareTo("class")==0 ) {
					Class cl;
					try {// .class on disk file occure to be a fake java class
						cl = Class.forName(Utils.getBase(f));
					} catch (Exception e) {
						//log("Failed loading " + f.getName() + " with msg: " + e.getMessage());
						continue;
					}
					if (PeriodicSignal.class.isAssignableFrom(cl) && Utils.isInstantiable(cl)) { // loaded class is a subclass of PeriodicSignal
						
						JRadioButton rb = new JRadioButton(cl.getName());
				    	signalButtonGroup.add(rb);
						rb.addActionListener(signalButtonListener);
						signalButtonsPanel.add(rb);
					}
				}
			}
			
			// activate the first button
			Enumeration e = signalButtonGroup.getElements();
			if (!e.hasMoreElements()) {
				log("No signals found");
				System.exit(0);
			}
			//((JRadioButton) e.nextElement()).setSelected(true);
			((JRadioButton) e.nextElement()).doClick(); // start playing first
			
			*/
			
			
			
			
			
			
			dialog.pack();
			dialog.setVisible(true);
			
/*			log ("press Contorl-C to terminate");
			while (true) {// halt playing 
				Thread.sleep(3000);
				
				//log("reopening");
				
				// once closed it seems not possible to re-open the line
				//line.close(); // what if fails then we'll not be able to nullify the reference?
				//line.open(audioFormat, table, 0, table.length);
				
				
				//line.close(); // what if fails then we'll not be able to nullify the reference?
				//line = null;
				//obtainAndOpenLine(); // does not work either - no errors, no sound
			}
			
*/			
			dialog.setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
			
		} finally {
			if (line != null) // what if line.close fails in the block?
				//line.close(); //close on window close
				;
		}
		
	}

	// prints some info about availble device capabilities
/*	private static void printLineInfos(Line.Info[] infos) {
		log("got " + infos.length + " line infos:");
		
		for (int i = 0; i < infos.length; i++) {
			if (infos[i] instanceof DataLine.Info) {
		    	DataLine.Info dataLineInfo = (DataLine.Info) infos[i];
		    	AudioFormat[] supportedFormats = dataLineInfo.getFormats();
		    	log("  [" + i + "] " + dataLineInfo + ", supported formats:");
		    	for (int j = 0; j < supportedFormats.length; j++)
		    		log("    (" + j + ") " +  supportedFormats[j]);
		  	} else
		  		log ("  [" + i + "] wrong info object class - " + infos[i].getClass().getName());
		}		
	}
*/	
	private static void printUsageAndExit() {
		log("OscillatorPlayer: usage:");
		log("\tjava OscillatorPlayer [-o <waveformtype>] [-a <amplitude>] [-f <signalfrequency>] [-r <samplerate>]");
		log("\tamplitude: 0..1");
		log("\toscillator types: Sine, Square, Triangle, Sawtooth");
		System.exit(1);
	}


	private static void log(String strMessage) {
		System.out.println(strMessage);
	}
}

