// Valentin Tihomirov, 2006, www.abelectron.com

/*This player does not use the javax.sound error-prone Line controls. The data that are provided by input stream are copied from signal 
waveform table. The content of this table is updated by controls. This method has additional feature of not requireing to stop playback 
on updates and, consequently, allows for smooth sound shifts.*/

// If it is hard to grip for you what is going on here, consider the non-gui version first.

/* TODO: 
 - Replace the table of samples by continous sampling. Current frequencies are not stable 
 when looknog at oscilloscope due to the buffer limitation.
 
*/


import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;

public class Player {
	
	public static void log(String msg) {
		Gui.log(msg);
	}

	// Gui-controlled vars.
	static float amplitude;
	static PeriodicSignal signal;
	static float signalFrequency;
	
	
	// Gui events handlers	
	public static void windowClosed() {in.remainingFrames = 0;}
	public static void setDutyCycle(float dc) {((DutyCycle) signal).setDutyCycle(dc);}
	public static void setSignal(PeriodicSignal signal) {Player.signal = signal;}
	public static void onAmplitudeChange() {updateTable();}
	public static void onFrequencyChange() {in.setWaveformTable(createNewTable());}
	
	// the options exposed by Gui
	public static String options = "";
	public static String optionDescription = "";

	// vars	used by static vars
	static byte[] table;
	static OscillatorInputStream in;
	static int framesPerPeriod;

	// called at input stream initializatin and when signal slider or signal waveform is changed by gui
	private static byte[] createNewTable() {
		framesPerPeriod = Math.round(Gui.line.getFormat().getFrameRate() / signalFrequency); // length of one period in frames
		//log("new table, frames per period = " + framesPerPeriod);
		int tableLength = framesPerPeriod * Gui.line.getFormat().getFrameSize();
		table = new byte[tableLength];
		updateTable();
		return table;
	}
	
	// called by createNewTable and when amplitude or duty cycle is changed by gui
	private static void updateTable() {
		// It would be nice to synchronize with "in" object for cache coherency. Otherwise, 
		// the reader thread may overlook the data modified by gui thread. However, 
		// for the first time this method is called by before the input stream is created.
		// synchronized (in) {updateTable();};
		
		float a = (float) (amplitude * Math.pow(2, Gui.line.getFormat().getSampleSizeInBits()-1) -1);
		for (int frame = 0; frame < framesPerPeriod; frame++) {
			
			float periodPosition = (float) frame / (float) framesPerPeriod;
			short sample = (short) Math.round(signal.getValue(periodPosition) * a);
			
			int baseAddr = frame * Gui.line.getFormat().getFrameSize();
			
			for (int ch=0; ch < Gui.line.getFormat().getChannels(); ch++) {
				table[baseAddr++] =  (byte) (sample & 0xFF);
				if (Gui.line.getFormat().getSampleSizeInBits() == 16)
					table[baseAddr++] = (byte) ((sample >>> 8) & 0xFF);
			}
		}
	}

	
	public static void pump() throws java.io.IOException {
	
		float responseTime = 1; // time in seconds. Must translate from 0 to max (or 1 to max in log scale) in this time.
		

		
			in = new OscillatorInputStream(Gui.line.getFormat(), createNewTable(), AudioSystem.NOT_SPECIFIED);
			
			int	numBytesRead;
			while ((numBytesRead = in.read(Gui.buf)) != -1)
				Gui.bufToLine(numBytesRead);
				
	}

	
}


