/*This file describes a period of a signal wafeform. In fact, the getValue must be static (since no state info is stored); nevertheless, static methods are not inheritable. Such system allows for custom extentions of oscillators.*/

/**	The relative position inside the period
	of the waveform. 0.0 = beginning, 1.0 = end
*/

/*returns a value [-1.0..1.0] at the moment of time t inside the period [0.0..1.0] of the waveform*/
interface PeriodicSignal {
	float getValue(float t);
}


/*Signed signals range values from -1 to 1*/

class Sine implements PeriodicSignal {
	public float getValue(float t) { return (float) Math.sin(t * 2.0 * Math.PI);};
}

class Square implements PeriodicSignal {
	public float getValue(float t) {return (t < 0.5 ? 1.0f : -1.0f);};
}

class Triangle implements PeriodicSignal {
	public float getValue(float t) {return (t < 0.5 ? 4*t-1 : 3-4*t);};
}

class Saw implements PeriodicSignal {
	public float getValue(float t) {return 2*t-1;};
}

class Circle implements PeriodicSignal {
	// (x-r)^2+y^2=1  =>  y = sqrt(x*(2r-x)), x <2r, r=.25
	public float getValue(float x) { 
		
		if (x < .5)
			return (float) Math.sqrt(x*(.5-x))*4; // scale .25 amplitudes 4x to max = 1
		else {
			x -= .5;
			return (float) -Math.sqrt(x*(.5-x))*4;
		}
			
	};
}


