/* TODO: try with imageCanvas. 
	public void blitWorkplaceToScreen(Rectangle clip) {
		Graphics screenGraphics = container.getGraphics();
		if(clip != null) {
			screenGraphics.setClip(clip);
		}
		screenGraphics.drawImage(wpBuffer, 0, 0, container);
		screenGraphics.dispose();
	}
	
Problems: volatile image is not reliable	
	*/
	
	

import java.awt.*;

/**Minimizes rendering by storing it into a buffer and supplies the cached image to restore clipped regions. Eliminated flickering naturally. */
public abstract class ImageComponent extends Component { // Should we derive from Canvas? Component? Container?
	Image cache = null;
	
	/**Invalidates the cache*/
	public void invalidate() {
		cache = null;
		super.invalidate(); // do we need call super?
	}
		
	
	/*Descedants must override this method as a replacement of paint().*/
	abstract public void render(Graphics g);

	/**Inastantiate the buffer and render graphics on it.*/
	protected void createImage() {
		
		Dimension d = getSize();
		
		// get a double buffer of the component
		// May be there is no need copy the image and use existing some buffer instead?
			
		
		//cache = createImage(d.width, d.height); // about 64 ms for 1000x1000 bitmap
		
		cache = createVolatileImage(d.width, d.height); // about 3ms for 1000x1000 bitmap
		
		Graphics gg = cache.getGraphics();
		render(gg);	// render into the buffer, about 1.5ms for 1000x1000 bitmap	
		gg.dispose();
		
	}
	
	/**Repaints the buffer without instantiating a new. Used by descedants*/
	protected void repaintImage() {
		Graphics g = cache.getGraphics();
		g.clearRect(0,0, getWidth(), getHeight());
		render(g);
		g.dispose();
	}
	
	
	/** Supplies the clipped area from the region of cache*/
	protected void paintFromBuffer(Graphics g) {
		Rectangle clip = g.getClipBounds();
		int x1 = clip.x;
		int x2 = clip.x+clip.width;
		int y1 = clip.y;
		int y2 = clip.y+clip.height;
		//g.drawImage(cache, x1, y1, x2, y2, x1, y1, x2, y2,this); // result is always true
		g.drawImage(cache, x1, y1, x2, y2, x1, y1, x2, y2, null); // result is always true, takes about 3ms for 1000x1000 bitmap
	}

	/**Render into the buffer if nothing is cached and restore the clipped area from it.*/
	public void paint(Graphics g) {
		
		if (cache == null)
			createImage();
			
		paintFromBuffer(g);
		
	}
	
	/**The image size is not changed on app updates; therefore, this method updates the buffer without allocations of a new one.*/
	public void repaint() { 
		invalidate();
		super.repaint();
	}
		
	public boolean isOpaque() {return true;} // do we need this?

	private static void log(String strMessage) {
		System.out.println(strMessage);
	}
	
}


/*	

		// performance monitor
		long time = System.currentTimeMillis();
		for (int i = 0; i < 100; i++) {
		}
		System.out.println(System.currentTimeMillis() - time);
*/	








/**Renders in repaint() rather than disposes the buffer.*/
abstract class ImageComponent2 extends ImageComponent { 
	
	/**The image size is not changed on app updates; therefore, this method updates the buffer without allocations of a new one.*/
	public void repaint() { 
		if (cache != null)
			repaintImage(); // update the buffer
			
		super.repaint();
	}
	
}






/**Activates the repaint flag in the repaint() deferring the rendering to paint()*/
abstract class ImageComponent3 extends ImageComponent {
	boolean repaint; // no need for initialization since the null cache overrides this flag
	
	/*Descedants must override this method as a replacement of paint().*/
	abstract public void render(Graphics g);
	
	/**Render into the buffer if nothing is cached and restore the clipped area from it.*/
	public void paint(Graphics g) {
		if (cache == null) {
			createImage();
			repaint = false;
		}
		
		if (repaint) { // re-render the image without reallocating the buffer
			repaintImage();
			repaint = false;
		}
		
		paintFromBuffer(g);
		
	}
	
	/**The image size is not changed on app updates; therefore, this method updates the buffer without allocations of a new one.*/
	public void repaint() { 
		repaint = true;
		super.repaint();
	}
		
}


