// Run it in a separate thread

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.GeneralPath;
import javax.swing.*;
import javax.swing.event.*;

class WaveformVisualizer extends Canvas {
	//static int WIDTH = 500;
	static int HEIGHT = 257;
	static byte[] bytes;
	static boolean unsigned = false;
	
	public void paint(Graphics g) {getBufferStrategy().show();}
	
	int offset = 0;
	public void render() {

		Graphics g = getBufferStrategy().getDrawGraphics();
		try {
			final int w = getWidth();
			final int h = getHeight();
			g.clearRect(0,0,getWidth(),h);
			
			GeneralPath gp = new GeneralPath();
			if (unsigned) {
				gp.moveTo(0, 255 - bytes[offset] + 1); // unsigned
				for (int x = 1, pos = offset+1; x < w; x++, pos++) //iterate the screen from left to right
					gp.lineTo(x, 255 - (((int) bytes[pos]) & 0xFF) + 1);
			} else { 
				gp.moveTo(0, 127 - bytes[offset] + 1); // signed
				for (int x = 1, pos = offset+1; x < w; x++, pos++) //iterate the screen from left to right
					gp.lineTo(x, 127 - bytes[pos] + 1);
			}
	        ((Graphics2D) g).draw(gp);
		} finally {
			g.dispose();
		}
	}
	
	WaveformVisualizer() {
		//setBounds(0,0,WIDTH,HEIGHT);
		//setBounds(0,0,bytes.length, 255);
		//setPreferredSize(new Dimension(bytes.length, 255));
		setSize(800, HEIGHT);
	}
	

	public static void main(String args[]) throws Exception {
        
		if (args.length < 1) {
			log("A 8bit mono file parameter expected. The 2nd 'unsigned' param is optional, 'signed' samples are default.");
			return;
		}
		bytes = Utils.getBytesFromFile(args[0]);
		if (args.length > 1 && args[1].compareToIgnoreCase ("unsigned") == 0)
	        unsigned = true;
			
		/*StringBuffer sb = new StringBuffer(bytes.length*4);
		for (int x = 0; x < bytes.length; x++) {
			int i = bytes[x] & 0xFF;
			sb.append(i).append(" ");
		}
		log(sb.toString());
		*/
		final JFrame frame = new JFrame("Buffer Visualiser");
		Container panel = frame.getContentPane();
		
		//panel.setPreferredSize(new Dimension(WIDTH,HEIGHT));
		//panel.setLayout(null);
		//panel.add(new JScrollPane(new BufferVisualiserWindow()), BorderLayout.CENTER);
		final WaveformVisualizer canvas = new WaveformVisualizer();
		panel.add(canvas, BorderLayout.CENTER);
		//panel.add(new BufferVisualiserWindow(), BorderLayout.CENTER);
		log(bytes.length + (unsigned ? " un" : " ") + "signed samples, max offset = " + (bytes.length - canvas.getWidth()));
		JScrollBar sb = new JScrollBar(JScrollBar.HORIZONTAL, 0, 50, 0, bytes.length - canvas.getWidth());
		sb.setBlockIncrement(100);
		sb.setUnitIncrement(20);
		sb.requestFocusInWindow();
		//final Insets ins = panel.getInsets();
		sb.addAdjustmentListener(new AdjustmentListener() {
			public void adjustmentValueChanged(AdjustmentEvent e) {
				//canvas.setLocation(-e.getValue(), 0);
				frame.setTitle("Offset = " + e.getValue());
				if (canvas.offset != e.getValue()) {
					canvas.offset = e.getValue(); 
					canvas.render();
					canvas.repaint();
				}
			}
		} ) ;
		panel.add(sb, BorderLayout.SOUTH);
		
		
		canvas.addMouseMotionListener(new MouseMotionListener() {
			public void mouseDragged(MouseEvent e) {}
			public void mouseMoved(MouseEvent e) {
				//frame.setTitle("mouse at " + e.getX() + ", " + e.getY());
				int pos = canvas.offset + e.getX();
				frame.setTitle("Samples[" + pos + "] = " + (unsigned ? (((int) bytes[pos]) & 0xFF) : bytes[pos]));
				
		     }
		});
		
		//frame.setBounds(0,0,WIDTH,HEIGHT);
		frame.pack();
		frame.setVisible(true);
		frame.addWindowListener( new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				System.exit(0);
			}
		});
		frame.setResizable(false);
		canvas.createBufferStrategy(2);
		canvas.render();
		canvas.repaint();
		
	}
	
	static void log(String strMessage) {
		System.out.println(strMessage);
	}
	
}

