
import javax.swing.*;
import javax.swing.event.*;
import java.io.Serializable;
import java.util.EventListener; 

// should it be monotonic? User must read float value, the integer one is for 
// slider only.

// be sure that "int -> float -> int" translation does not change the int. Otherwise, you are prevented 
// from advancing the JSlider by arrorw keys.
public class IntFloatBoundedRangeAdapter implements BoundedRangeModel, Serializable, ChangeListener {

    protected transient ChangeEvent changeEvent = null;
    protected EventListenerList listenerList = new EventListenerList();

    protected FloatBoundedRangeModel floatModel = null;
    protected int min;
    protected int max;
    private float k;
    
    public IntFloatBoundedRangeAdapter(FloatBoundedRangeModel model, int grades) {
	    this(model, 0, grades-1);
    }
    
    public IntFloatBoundedRangeAdapter(FloatBoundedRangeModel model, int min, int max) {
	    this.min = min;
	    this.max = max;
	    this.floatModel = model;
	    setFloatModel(model);
    }

    /*Convinience method*/
    public void setFloatValue(float value) {
	    floatModel.setValue(value);
    }

    /*Convinience method*/
    public float getFloatValue() {
	    return floatModel.getValue();
    }
    
        
    private void updateK() {
		k = (floatModel.getMaximum() - floatModel.getMinimum())/(getMaximum() - getMinimum());
    }
    
    /** 
     * Refresh k on bounds change. It will be a problem here once float model 
     * borders change and its another listener accessing this model is invoked
     * prior to this handler.
     */
    public void stateChanged(ChangeEvent event) {
	    updateK();
    }
    
    public void setFloatModel(FloatBoundedRangeModel model) {
	    removeChangeListener(this);
	    this.floatModel = model;
	    updateK();
	    addChangeListener(this);
    }
    
    public FloatBoundedRangeModel getFloatModel() {
	    return floatModel;
    }
    
	/**The integer model values are translated to/from the floating values stored.*/
	public int getMinimum() {return min;}
	public void setMinimum(int min) {
		
// 		//throw new RuntimeException("Prohibited operation, sorry for Liskov insubstituriability");
// 		
// 		//before: min = 0, max = grades-1, grades = max - min in total
// 		//after: min = newMinimum, max = grades-1, max - min = grades - newMin in total
// 		grades -= minimum;
// 		if (minimum != 0)
// 			floatModel.fireStateChanged();
		setRangeProperties(Math.max(getValue(), min), 0, min, getMaximum(), getValueIsAdjusting());

	}
	
	public int getMaximum() {
		return max;
	}
	public void setMaximum(int max) {
		setRangeProperties(Math.min(getValue(), max), 0, 0, max, getValueIsAdjusting());
// 		if (grades != newMaximum+1) {
// 			grades = newMaximum+1;
// 			floatModel.fireStateChanged();
// 		}
	}
	
	float translate(int value) {
		return floatModel.getMinimum() + k * (value - min);
	}
	
	int translate(float value) {
		return min + (int) Math.round((value - floatModel.getMinimum())/k) ;
	}
	
	
	
	public int getValue() {
		return translate(floatModel.getValue());
	}
		
	public void setValue(int value) {
		setRangeProperties(Math.max(Math.min(value, max), 0), 0, getMinimum(), getMaximum(), getValueIsAdjusting());
	}
    
	
	
    public int getExtent() {return 0;}
    public void setExtent(int newExtent) {} // do nothing

    

    public void setValueIsAdjusting(boolean b) {
	    floatModel.setValueIsAdjusting(b);
    }

    public boolean getValueIsAdjusting() {
	    return floatModel.getValueIsAdjusting();
    }

    
    
    public void setRangeProperties(int newValue, int extent, int newMin, int newMax, boolean adjusting) {
		if (newMin > newMax) {
			newMin = newMax;
		}
		
		// why do they prefer adjusting the max?
		if (newValue > newMax) {
			newMax = newValue;
		}
		
		if (newValue < newMin) {
			newMin = newValue;
		}

				
// 		newValue -= newMin;
// 		newMax -= newMin;
// 		// newMin = 0;
// 		
//         boolean isChange =
//             (newValue != getValue()) ||
//             (newMax != grades-1) ||
//             adjusting != getValueIsAdjusting();

// 		if (isChange) {
//             grades = newMax+1;
//             floatModel.setRangeProperties(translate(newValue), 0, floatModel.getMinimum(), floatModel.getMaximum(), adjusting);
//         }

		if (newMin != min || newMax != max) {
		
			// adjust borders
			max = newMax;
			min = newMin;
			
			//force the event triggering
			floatModel.setRangeProperties(translate(newValue), 0, floatModel.getMinimum(), floatModel.getMaximum(), !adjusting);
		
		}
		
		floatModel.setRangeProperties(translate(newValue), 0, floatModel.getMinimum(), floatModel.getMaximum(), adjusting);
		
	}
	
	public void addChangeListener(ChangeListener l) {
		floatModel.addChangeListener(l);
	}
	
	
	public void removeChangeListener(ChangeListener l) {
		floatModel.removeChangeListener(l);
	}
	

//     public ChangeListener[] getChangeListeners() {
//         return floatModel.getChangeListeners();
//     }

    
    /**
     * Returns a string that displays all of the 
     * <code>BoundedRangeModel</code> properties.
     */
    public String toString()  {
        String modelString =
        	getValue() + " in (" + min + ", " + max + "), " + 
            "floatModel=[" + floatModel + "]";

        return getClass().getName() + "[" + modelString + "]";
    }

//     public <T extends EventListener> T[] getListeners(Class<T> listenerType) { 
// 		return floatModel.getListeners(listenerType); 
//     }
    
}



