/*

	
*/

//package titlenter;



import java.awt.*; 
import java.awt.event.*; 
import javax.swing.*;
import javax.swing.event.*;

/** 
 * Enables editing floating value of the slider class defined in the frame in it's title 
 * bar.
 */
public class FloatTitleEditFrame extends JFrame { 
	

	
	/** Restarted when user types a key in edit mode. Exits the edit mode on elapse. 
	 * Running timer is prematurely terminated when title is updated by anything 
	 * other than edit.
	 **/
	final Timer timer;

	/** 
	 * Switches between key adapters accepting int and float parts of the value 
	 * being entered by user. The first digit is accepted by init adaper to enter 
	 * the mode.
	 **/
	private void switchToKeyListener(KeyListener replacement) {
		
// 		ToDo: switchKeyListener relies on getKeyListeners()[0]. Make sure there is no potential 
// 		errors. Itinitlly, there should not, since we create the slider and add the first 
// 		listener. However, later things may change and we remove and add the listener to the 
// 		end. Currently, I see two solutions: ether remove all the key listeners 
// 		(init, int, float) in the switchToKeyListener before adding new, or keep current
// 		listener in a separate Slider's field. The move the switch method into the Slider
// 		class then.

		//editSlider.removeKeyListener(editSlider.getKeyListeners()[0]);
		
		// We do not know which is the current mode, so remove all for sure.
		editSlider.removeKeyListener(intKeyAdapter);
		editSlider.removeKeyListener(initKeyAdapter);
		editSlider.removeKeyListener(floatKeyAdapter);
		
		editSlider.addKeyListener(replacement);
	}
	
	/** Overrieaden to cease the edit mode.*/
	public void setTitle(String title) { 
		super.setTitle(title);
		//System.out.println(title + ", timer is running = " + timer.isRunning());
		if (timer.isRunning()) { // stop ongoing edit
			timer.stop();
			switchToKeyListener(initKeyAdapter);
			//editSlider.setValueIsAdjusting(false);
		}
	}
	
	/** Appends a digit to title in edit mode*/
	private void appendTitle(char c) {
		super.setTitle(getTitle() + c);
		timer.restart();
	}

	
	
	/** 
	 * Allows to enter precise value manually by typing digits on focused slider. 
	 * The onFocus event is added for convince. Feel free to 
	 * removeFocusListener(sliderFocusAction).
	 * Implement the abstract methods like in Template Pattern.
	 **/
	public abstract class Slider extends JSlider {
		
		protected float value; // used by manual enter
		
// 		public FloatBoundedRangeModel getFloatModel() {
// 			//return (FloatBoundedRangeModel) getModel();
// 			return ((IntFloatBoundedRangeAdapter) getModel()).getFloatModel();
// 		}
		
		public float getFloatValue() {
			//return getFloatModel().getValue();
			return ((IntFloatBoundedRangeAdapter) getModel()).getFloatValue();
		}
		
		public void setFloatValue(float value) {
			this.value = value;
			
			// prevent title corruption on slider update
			removeChangeListener(sliderChangeListener);
			//getFloatModel().setValue(value); 
			((IntFloatBoundedRangeAdapter) getModel()).setFloatValue(value);
			addChangeListener(sliderChangeListener);
		}
		
		/** Appears on title when entering edit mode.*/
		public String editPrefix;
		
		// used to show value on title bar when user slides and timer terminates the edit mode
		abstract public String formatValue();
		
		public Slider(IntFloatBoundedRangeAdapter model, String prefix) {
			super(model);
			this.editPrefix = prefix;
			
			addChangeListener(sliderChangeListener);
			
			addFocusListener(sliderFocusAction);
			addKeyListener(initKeyAdapter);
			
		}
		
	
		/**Brings formatValue() to the form's title.*/
		private void showFormattedValue() {
			//setTitle("Freq = " + freq + " Hz");
			setTitle(formatValue());
		}	
		
	}
	
	class DefaultFormatSlider extends Slider {
		public String magnitude;
		public String unit;
		DefaultFormatSlider(IntFloatBoundedRangeAdapter model, String magnitude, String unit) {
			super(model, "Manual " + magnitude + " = ");
			this.magnitude = magnitude;
			this.unit = unit;
		}
		
		public String formatValue() {
			return magnitude + " = " + getFloatValue() + " " + unit;
		}
		
	}
	
	/** The slider currently being edited. Initiated by initKeyLiseter*/
	private Slider editSlider;
	
	/**
	 * Processes user sliding by updating the title bar and float value. Since it is 
	 * the first listener in the list, I beleive it will be fired first, so you can 
	 * read correct values from getFloatValue(). Peahaps I'd better move the updates
	 * into a FloatRange model.
	 */
	private final ChangeListener sliderChangeListener = new ChangeListener() {
		public void stateChanged(ChangeEvent e) {
			((Slider) e.getSource()).showFormattedValue();
		}
	};
	
	/**Updates the title bar.*/
	public final FocusListener sliderFocusAction = new FocusAdapter() {
		public void focusGained(FocusEvent e) {
			((Slider) e.getComponent()).showFormattedValue();
		}
	};
	
	
	
	
	
	//
	// Key listeners
	//
	/**Filters out typed digits for in[i]t key adapter. */
	private abstract class BasicDigitKeyAdapter extends KeyAdapter {

		char c;
		boolean dot;
					
		// invoked this method if filter is passed. c and dot fields are valid on 
		// the call.
		abstract void onDigit(KeyEvent e);
		
 		public void keyTyped(KeyEvent e) {
	 		
			c = e.getKeyChar();
			dot = (c == '.');
			if (Character.isDigit(c) || dot )
				onDigit(e);

		}
	} 
	
	/**Accepts fractional part of the float value showing it on the title and firing 
	change events.*/
	private final FloatKeyAdapter floatKeyAdapter = new FloatKeyAdapter();
	private class FloatKeyAdapter extends KeyAdapter {
		float multiplier;
 		public void keyTyped(KeyEvent e) {
			char c = e.getKeyChar();
			if (Character.isDigit(c)) {
				//System.out.println("float");
				
				editSlider.setFloatValue(editSlider.value + Character.digit(c, 10) * multiplier);
				
				multiplier *= .1;
				appendTitle(c);
				
			}
		}
	}; // floatKeyAdapter
	

	/** Accepts leading part of the float value showing it on the title and firing 
	change events.*/
	private final KeyListener intKeyAdapter = new BasicDigitKeyAdapter() { // intKeyAdapter
		
 		public void onDigit(KeyEvent e) {
			//System.out.println("int");
			
			if (dot) {
				switchToKeyListener(floatKeyAdapter);
				floatKeyAdapter.multiplier = .1f;
			} else
				editSlider.setFloatValue(editSlider.value * 10 + Character.digit(c, 10));
				
			appendTitle(c);
			
		}
	}; // intKeyAdapter

	
	/** Resets the title to prefix while float value to 0 and passes all the further 
	processing to intKeyAdapter.*/
	final KeyListener initKeyAdapter = new BasicDigitKeyAdapter() { 
	
 		public void onDigit(KeyEvent e) {
			//System.out.println("init");

			Slider src = (Slider) e.getComponent();
			
			setTitle(src.editPrefix); // this will also terminate current edit
			
			editSlider = src;
			editSlider.value = 0;

						
						
			// process this and following digits by int listener
			//processKeyEvent(e); // re-fire  the event
			switchToKeyListener(intKeyAdapter);
			intKeyAdapter.keyTyped(e);
		}
	}; // initialKeyAdapter

	
	//
	// End of listeners section
	// 	
	


	
	/**Constructing with a title does not usually make sense, since it is overwritten by onfucus 
		of the first slider added.*/
	public FloatTitleEditFrame() {
		
		super();
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE); 

		timer = new Timer(3000, new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				//System.out.println("timer elapsed");
				
				// it was set up by key typing, just update the title bar and 
				// init key listener
				editSlider.showFormattedValue(); 
			}
		});
		
		// despite we need a single-shot timer, it must be running on elapsed event 
		// since it is the edit mode flag
		timer.setRepeats(true); 
	
	} 
	
}


