/*
ToDone: the inner JSlider has been replaced with a FloatModel. 

The loads of problems encured: 
	1)We cannot extend the float model similarily to JSlider, since a reference to model 
	implementation is needed. 
	2) The scroll bars do not have the change listeners. Despite we could accomplish via 
	model's change listener, the problem is deeper. The change, focus and key listeners 
	cannot be form's singletones, since the event source component's class is unknown and, 
	hence, cannot return the model. 
	3) The model must keep the reference to its component to enable keyListener switching.
	4) Finally, the design becomse redundant in the way that user must make two things to 
	enable editing: provide a model to the wripper and then to the control: 
		SLIDER.setModel(MODEL); 
		frame.new EMW(MODEL, prefix, SLIDER);
	It is easy to mistake here, leading to inconsistency.
	
	
	Due to the listed problems it seems impossible to accomplish 
	a good design for the good idea.
	
*/

import java.awt.*; 
import java.awt.event.*; 
import javax.swing.*;
import javax.swing.event.*;

/** 
 * Enables editing floating value of the slider class defined in the frame in it's title 
 * bar.
 */
class FloatTitleEditFrame extends JFrame { 
	

	
	/** Restarted when user types a key in edit mode. Exits the edit mode on elapse. 
	 * Running timer is prematurely terminated when title is updated by anything 
	 * other than edit.
	 **/
	final Timer timer;

	/** 
	 * Switches between key adapters accepting int and float parts of the value 
	 * being entered by user. The first digit is accepted by init adaper to enter 
	 * the mode.
	 **/
	private void switchToKeyListener(KeyListener replacement) {
		editingModel.component.removeKeyListener(editingModel.component.getKeyListeners()[0]);
		editingModel.component.addKeyListener(replacement);
	}
	
	/** Overrieaden to cease the edit mode.*/
	public void setTitle(String title) { 
		super.setTitle(title);
		//System.out.println(title + ", timer is running = " + timer.isRunning());
		if (timer.isRunning()) { // stop ongoing edit
			timer.stop();
			switchToKeyListener(editingModel.initKeyAdapter);
			//editSlider.setValueIsAdjusting(false);
		}
	}
		
	/** Appends a digit to title in edit mode*/
	private void appendTitle(char c) {
		super.setTitle(getTitle() + c);
		timer.restart();
	}

	
	
	/** 
	 * Allows to enter precise value manually by typing digits on focused slider. 
	 * The onFocus event is added for convince. Feel free to 
	 * removeFocusListener(sliderFocusAction).
	 * Implement the abstract methods like in Template Pattern.
	 **/
	abstract class EditableModelWrapper extends JSlider {
		
		final FloatBoundedRangeModel model;
		final Component component;
		 
		protected float value; // used by manual enter
		
		public void setFloatValue(float value) {
			this.value = value;
			
			// prevent title corruption on slider update
			model.removeChangeListener(sliderChangeListener);
			model.setFloatValue(value); 
			model.addChangeListener(sliderChangeListener);
		}
		
		/** Appears on title when entering edit mode.*/
		public String editPrefix;
		
		// used to show value on title bar when user slides and timer terminates the edit mode
		abstract public String formatValue();
		
		EditableModelWrapper(FloatBoundedRangeModel model, String prefix, Component c) {
			this.model = model;
			this.editPrefix = prefix;
			component = c;
			
			model.addChangeListener(sliderChangeListener);
			
			c.addFocusListener(sliderFocusAction);
			c.addKeyListener(initKeyAdapter);
			
		}
		
	
		/**Brings formatValue() to the form's title.*/
		private void showFormattedValue() {
			//setTitle("Freq = " + freq + " Hz");
			setTitle(formatValue());
		}	
		
		/**
		 * Processes user sliding by updating the title bar and float value. Since it is 
		 * the first listener in the list, I beleive it will be fired first, so you can 
		 * read correct values from getFloatValue(). Peahaps I'd better move the updates
		 * into a FloatRange model.
		 */
		private final ChangeListener sliderChangeListener = new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				showFormattedValue();
			}
		};
		
		/**Updates the title bar.*/
		public final FocusListener sliderFocusAction = new FocusAdapter() {
			public void focusGained(FocusEvent e) {
				showFormattedValue();
			}
		};

		
		
				
		
		//
		// Key listeners
		//
		/**Filters out typed digits for in[i]t key adapter. */
		private abstract class BasicDigitKeyAdapter extends KeyAdapter {
	
			char c;
			boolean dot;
						
			// invoked this method if filter is passed. c and dot fields are valid on 
			// the call.
			abstract void onDigit(KeyEvent e);
			
	 		public void keyTyped(KeyEvent e) {
		 		
				c = e.getKeyChar();
				dot = (c == '.');
				if (Character.isDigit(c) || dot )
					onDigit(e);
	
			}
		} 
		
		/**Accepts fractional part of the float value showing it on the title and firing 
		change events.*/
		private final FloatKeyAdapter floatKeyAdapter = new FloatKeyAdapter();
		private class FloatKeyAdapter extends KeyAdapter {
			float multiplier;
	 		public void keyTyped(KeyEvent e) {
				char c = e.getKeyChar();
				if (Character.isDigit(c)) {
					System.out.println("float");
					
					setFloatValue(editingModel.value + Character.digit(c, 10) * multiplier);
					
					multiplier *= .1;
					appendTitle(c);
					
				}
			}
		}; // floatKeyAdapter
		
	
		/** Accepts leading part of the float value showing it on the title and firing 
		change events.*/
		private final KeyListener intKeyAdapter = new BasicDigitKeyAdapter() { // intKeyAdapter
			
	 		public void onDigit(KeyEvent e) {
				System.out.println("int");
				
				if (dot) {
					switchToKeyListener(floatKeyAdapter);
					floatKeyAdapter.multiplier = .1f;
				} else
					setFloatValue(editingModel.value * 10 + Character.digit(c, 10));
					
				appendTitle(c);
				
			}
		}; // intKeyAdapter
	
		
		/** Resets the title to prefix while float value to 0 and passes all the further 
		processing to intKeyAdapter.*/
		final KeyListener initKeyAdapter = new BasicDigitKeyAdapter() { 
		
	 		public void onDigit(KeyEvent e) {
				System.out.println("init");
	
				setTitle(editPrefix); // this will also terminate current edit
				
				editingModel = EditableModelWrapper.this;
				value = 0;
							
				// process this and following digits by int listener
				//processKeyEvent(e); // re-fire  the event
				switchToKeyListener(intKeyAdapter);
				intKeyAdapter.keyTyped(e);
			}
		}; // initialKeyAdapter
	
		
		//
		// End of listeners section
		// 	
		
	}
	
	/** The slider currently being edited. Initiated by initKeyLiseter*/
	private EditableModelWrapper editingModel;
	
	
	FloatTitleEditFrame(String title) {
		
		super(title);
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE); 

		timer = new Timer(3000, new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				//System.out.println("timer elapsed");
				
				// it was set up by key typing, just update the title bar and 
				// init key listener
				editingModel.showFormattedValue(); 
			}
		});
		
		// despite we need a single-shot timer, it must be running on elapsed event 
		// since it is the edit mode flag
		timer.setRepeats(true); 
	
	} 
	
}


