/*
	
Todo: 
	- FloatBoundRangeModel for slider. There is a problem though for non-linear sliders 
		like logarithmic scale. For conventional model sliders there is no problem since 
		there is only one point of contror and user exponentates the slider's value in 
		onChange handler to	produce logarithmic output.
		
		Currently, we approximate the integer position by conventional rounding when user 
		enters float point in edit mode. This assumes the scale is linear.


*/

import java.awt.*; 
import java.awt.event.*; 
import javax.swing.*;
import javax.swing.event.*;

/** 
 * Enables editing floating value of the slider class defined in the frame in it's title 
 * bar.
 *
 * The problem is that changing fractional part does not change the integer part which 
 * prevents change events from propagation in the integer BoundedRangeModel.
 **/
class FloatTitleEditFrame extends JFrame { 
	

	
	/** Restarted when user types a key in edit mode. Exits the edit mode on elapse. 
	 * Running timer is prematurely terminated when title is updated by anything 
	 * other than edit.
	 **/
	final Timer timer;

	/** 
	 * Switches between key adapters accepting int and float parts of the value 
	 * being entered by user. The first digit is accepted by init adaper to enter 
	 * the mode.
	 **/
	private void switchToKeyListener(KeyListener replacement) {
		editSlider.removeKeyListener(editSlider.getKeyListeners()[0]);
		editSlider.addKeyListener(replacement);
	}
	
	/** Overrieaden to cease the edit mode.*/
	public void setTitle(String title) { 
		super.setTitle(title);
		//System.out.println(title + ", timer is running = " + timer.isRunning());
		if (timer.isRunning()) { // stop ongoing edit
			timer.stop();
			switchToKeyListener(initKeyAdapter);
			//editSlider.setValueIsAdjusting(false);
		}
	}
	
	/** Appends a digit to title in edit mode*/
	private void appendTitle(char c) {
		super.setTitle(getTitle() + c);
		timer.restart();
	}

	
	
	/** 
	 * Allows to enter precise value manually by typing digits on focused slider. 
	 * The onFocus event is added for convince. Feel free to 
	 * removeFocusListener(sliderFocusAction).
	 **/
	abstract class Slider extends JSlider {
		
		/**Can be obtained by getFloatValue().*/
		protected float value;
		
		/**The limit for the user not to exeed when typing.*/
		public float fMax = Float.MAX_VALUE; // can be limited by setMax()
		
		/** Appears on title when entering edit mode.*/
		public String editPrefix;
		
		/** This float value is in sync with integer getValue, but is more precise.*/
		public float getExactValue() {return value;}
		
		// used to show value on title bar when user slides and timer terminates the edit mode
		abstract public String formatValue();
		
		Slider(int min, int max, int value, float fMax, String prefix) {
			super(min, max, value);
			this.value = value;
			this.fMax = fMax;
			this.editPrefix = prefix;
			
			addChangeListener(sliderChangeListener);
			addFocusListener(sliderFocusAction);
			addKeyListener(initKeyAdapter);
			
		}
		
	
		/**Brings formatValue() to the form's title.*/
		private void showFormattedValue() {
			//setTitle("Freq = " + freq + " Hz");
			setTitle(formatValue());
		}	
		
		/**
		 * Updates the float value and slider's both int value and position via setValue().
		 * This invokes notification of all the change listenters as normally. Just 
		 * remember to get value by getFloatValue() instead of getValue().
		 **/
		boolean setExactValue(float value) {
			
			// exit if value exeeds limits
			if (value > fMax) 
				return false;
			
			final int n = Math.min(Math.round(this.value = value), getMaximum());
			System.out.println("round(" + value + ") = " + n/* + ", isAsjusting = " + getValueIsAdjusting()*/);
			
			// move slider to new position and fire change event
			removeChangeListener(sliderChangeListener);
			setValue(n); 
			addChangeListener(sliderChangeListener);
			
			
			return true;
		}
	}
	
	/** The slider currently being edited. Initiated by initKeyLiseter*/
	private Slider editSlider;
	
	/**
	 * Processes user sliding by updating the title bar and float value. Since it is 
	 * the first listener in the list, I beleive it will be fired first, so you can 
	 * read correct values from getFloatValue(). Peahaps I'd better move the updates
	 * into a FloatRange model.
	 */
	private final ChangeListener sliderChangeListener = new ChangeListener() {
		public void stateChanged(ChangeEvent e) {
			Slider src = (Slider) e.getSource();
			src.setExactValue(src.getValue());
			src.showFormattedValue();
		}
	};
	
	/**Updates the title bar.*/
	public final FocusListener sliderFocusAction = new FocusAdapter() {
		public void focusGained(FocusEvent e) {
			((Slider) e.getComponent()).showFormattedValue();
		}
	};
	
	
	
	
	
	//
	// Key listeners
	//
	/**Filters out typed digits for in[i]t key adapter. */
	abstract class BasicDigitKeyAdapter extends KeyAdapter {

		char c;
		boolean dot;
					
		// invoked this method if filter is passed. c and dot fields are valid on 
		// the call.
		abstract void onDigit(KeyEvent e);
		
		// override this method for filtering and firing onDigit only on appropriate chars
		//boolean isDigit() { return Character.isDigit(c) || (c == '.');}
		
 		public void keyTyped(KeyEvent e) {
	 		
			c = e.getKeyChar();
			dot = (c == '.');
			if (Character.isDigit(c) || dot )
				onDigit(e);

		}
	} 
	
	/**Accepts fractional part of the float value showing it on the title and firing 
	change events.*/
	private final FloatKeyAdapter floatKeyAdapter = new FloatKeyAdapter();
	private class FloatKeyAdapter extends KeyAdapter {
		float multiplier;
 		public void keyTyped(KeyEvent e) {
			char c = e.getKeyChar();
			if (Character.isDigit(c)) {
				System.out.println("float");
				
				if (editSlider.setExactValue(editSlider.value + Character.digit(c, 10) * multiplier)) {
					multiplier *= .1;
					appendTitle(c);
				}
			}
		}
	}; // floatKeyAdapter
	

	/** Accepts leading part of the float value showing it on the title and firing 
	change events.*/
	private final KeyListener intKeyAdapter = new BasicDigitKeyAdapter() { // intKeyAdapter
		
 		public void onDigit(KeyEvent e) {
			System.out.println("int");
			
			if (dot) {
				switchToKeyListener(floatKeyAdapter);
				floatKeyAdapter.multiplier = .1f;
			} else
				if (!editSlider.setExactValue(editSlider.value * 10 + Character.digit(c, 10)))
					return;
				
			appendTitle(c);
			
		}
	}; // intKeyAdapter

	
	/** Resets the title to prefix while float value to 0 and passes all the further 
	processing to intKeyAdapter.*/
	final KeyListener initKeyAdapter = new BasicDigitKeyAdapter() { 
	
 		public void onDigit(KeyEvent e) {
			System.out.println("init");

			Slider src = (Slider) e.getComponent();
			
			setTitle(src.editPrefix); // this will also terminate current exit
			
			editSlider = src;
			editSlider.value = 0;
			//editSlider.setValueIsAdjusting(true);
			
			
						
			// process this and following digits by int listener
			//processKeyEvent(e); // re-fire  the event
			switchToKeyListener(intKeyAdapter);
			intKeyAdapter.keyTyped(e);
		}
	}; // initialKeyAdapter

	
	//
	// End of listeners section
	// 	
	


	
	
		
	//private static void createAndShowGUI() {
	FloatTitleEditFrame(String title) {
		
		super(title);
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE); 

		timer = new Timer(3000, new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				//System.out.println("timer elapsed");
				
				// it was set up by key typing, just update the title bar and 
				// init key listener
				editSlider.showFormattedValue(); 
			}
		});
		
		// despite we need a single-shot timer, it must be running on elapsed event 
		// since it is the edit mode flag
		timer.setRepeats(true); 
	
	} 
	
}



