/*
There are two opportunities to enter some text into title:
(1) invisible text field at the title. It acts like a dialog box requires 
	Enter or Escape to exit.
(2) Like in remote control, a timer starts when first digit of the (channel) 
	number is pressed. The timer resets on every new digit. No corrections 
	of the text is allowed. Once the timer stops, the new value can be entered.
	
Todo: 
	- pack the code into a SliderTitleEditFrame.
	- FloatBoundRangeModel for slider
	- oppose state machine (single key adapter with switch cases) to key processors listener sequence

*/

import java.awt.*; 
import java.awt.event.*; 
import javax.swing.*;
import javax.swing.event.*;

class Gui extends JFrame { 
	
	/*boolean editMode = false;
	boolean dotted = false; // when in editmode, shows whether number is int or float
	float freq;
	

	// ceases the frequency edit mode
	public void setTitle(String title) {
		//throw new UnsupportedOperationException("The setTitle() method is forbidden"); //throw new InvalidOpreationException
		setTitle(title, false);
	}

	
	
	// If timer is reset we enter/continue the edit mode. Otherwise, we
	// exit the edit mode stopping the timer.
	public void setTitle(String title, boolean resetTimer) {
		super.setTitle(title);
		editMode = resetTimer;
		dotted &= editMode; // reset to int mode when escapes the edit mode
	}
	
	*/
	
	// Exits the edit mode on elapse. Running timer is prematurely terminated
	// when title is updated by any controls.
	final Timer timer;

	float frequency;
	final JSlider sl1;
	
	void switchToKeyListener(KeyListener replacement) {
			sl1.removeKeyListener(sl1.getKeyListeners()[0]);
			sl1.addKeyListener(replacement);
	}
	
	// appends a digit to title in edit mode
	void appendTitle(char c) {
		setTitle(getTitle() + c, true);
	}
	
	// exits the edit mode
	public void setTitle(String title) { setTitle(title, false); }
	
	// we reset timer when are typing in edit mode. Otherwise, we
	// exit the edit mode stopping the timer.
	public void setTitle(String title, boolean editMode) {
		super.setTitle(title);
		if (!editMode) {
			timer.stop();
			switchToKeyListener(initKeyAdapter);
		} else
			timer.restart();
	}
	

	
	
	
	
	//
	// Key listeners
	//
	
	abstract class BasicDigitKeyAdapter extends KeyAdapter {

		char c;
		boolean dot;
					
		// invoked this method if filter is passed. c and dot fields are valid on 
		// the call.
		abstract void onDigit(KeyEvent e);
		
		// override this method for filtering and firing onDigit only on appropriate chars
		//boolean isDigit() { return Character.isDigit(c) || (c == '.');}
		
 		public void keyTyped(KeyEvent e) {
	 		
			c = e.getKeyChar();
			dot = (c == '.');
			if (Character.isDigit(c) || dot )
				onDigit(e);

		}
	} 
	
	// accepts fractional part of the frequency
	class FloatKeyAdapter extends KeyAdapter { // floatKeyAdapter
		float multiplier;
 		public void keyTyped(KeyEvent e) {
			char c = e.getKeyChar();
			if (Character.isDigit(c)) {
				System.out.println("float");
				setExactFrequency(frequency + Character.digit(c, 10) * multiplier);
				multiplier *= .1;
				appendTitle(c);
			}
		}
	}; // floatKeyAdapter
	final FloatKeyAdapter floatKeyAdapter = new FloatKeyAdapter();

	// accepts leading part of the frequency
	final KeyListener intKeyAdapter = new BasicDigitKeyAdapter() { // intKeyAdapter
		
 		public void onDigit(KeyEvent e) {
			System.out.println("int");
			if (dot) {
				switchToKeyListener(floatKeyAdapter);
				floatKeyAdapter.multiplier = .1f;
			} else
				setExactFrequency(frequency*10 + Character.digit(c, 10));
				
			appendTitle(c);
			
		}
	}; // intKeyAdapter

	// resets the title and freq
	final KeyListener initKeyAdapter = new BasicDigitKeyAdapter() { 
	
 		public void onDigit(KeyEvent e) {
			System.out.println("init");
			frequency = 0;
			setTitle("Manual freq = ", true);
			switchToKeyListener(intKeyAdapter);
			
			// process the digit by int listener
			//processKeyEvent(e); // re-fire  the event
			intKeyAdapter.keyTyped(e);
		}
	}; // initialKeyAdapter

	
	//
	// End of listeners section
	// 	
	
	// on slider will show its value, on timer - frequency
	void showFrequency(float freq) {
		setTitle("Freq = " + freq + " Hz");
	}
	
	// updates frequency variable and slider
	void setExactFrequency(float freq) {
		ChangeListener cl = sl1.getChangeListeners()[0];
		sl1.removeChangeListener(cl); // suspend listening
		
		int n = Math.min(Math.round(frequency = freq), sl1.getMaximum());
		System.out.println("round(" + frequency + ") = " + n);
		sl1.setValue(n);
		
		sl1.addChangeListener(cl); // restore listening
	}
	
	//private static void createAndShowGUI() {
	Gui() {
		
		super("TopLevelDemo");
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE); 

		sl1 = new JSlider(JSlider.HORIZONTAL, 0, 20000, 440);
		sl1.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				showFrequency(sl1.getValue());
			}
		});
		sl1.addFocusListener(new FocusAdapter() {
			public void focusGained(FocusEvent e) {
				showFrequency(sl1.getValue());
			}
		});
		getContentPane().add(sl1, BorderLayout.NORTH);
		
		/*sl1.addKeyListener(	new KeyAdapter() {
		      	
		 		public void keyTyped(KeyEvent e) {
			 		//frame.setTitle("Typed " + e.getKeyChar());
			 		
			 		char c = e.getKeyChar();
			 		boolean dot = (c == '.');
			 		if (Character.isDigit(c) || (!dotted && dot) ) {
				 		String title;
				 		if (!editMode) {
					 		frequency = 
				 		}
				 		
				 		frequency = editMode ? : 0;
				 		setTitle((editMode ? getTitle() : "") + c, true);
				 		dotted |= dot; // enter float mode on dot
			 		}
		      	}
		});*/
		
		

		sl1.addKeyListener(initKeyAdapter);
		
		
		final JSlider sl2 = new JSlider(JSlider.HORIZONTAL);
		getContentPane().add(sl2, BorderLayout.CENTER); 
		sl2.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				setTitle("ampl = " + sl2.getValue());
			}
		});
		
		timer = new Timer(3000, new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				System.out.println("timer elapsed");
				
				// it was set up by keys, just reset the title bar and 
				// init key listener
				showFrequency(frequency); 
			}
		});
		timer.setRepeats(false);
	
		//Display the window. 
		pack(); 
		setVisible(true);
	} 
	
	
	public static void main(String[] args) {
		
		//Schedule a job for the event-dispatching thread: //creating and showing this application's GUI. 
		javax.swing.SwingUtilities.invokeLater(new Runnable() { 
			public void run() { 
				//createAndShowGUI(); 
				new Gui();
			} 
		}); 
	}
}