
/*
Maps between integer ordinate and float value: fVal = offset + factor * base^iVal.
The integers are limited to bounded range interval. Exploiting the exponential feature,
where b^(dx+int) = b^dx * b^int = f * b^int, one can afford fo shift the bounded range 
interval of the ordinate itno floating domain. However, the interval length and 
discretness remains unchanged, so that shifting left int border to a desired float 
ordinate (the exponent), the right border will be shifted the same amount.

The escape would be a 3-layer architecture:
	integer BRM < = (int-float adapter) = > float RM ordinate < = (log or another) function = > float value
	
The function would be a FloatBoundedRangeModel with a reference to ordinate 
FloatBoundedRangeModel. The Int-Float adapter would refer to function for direct control 
the value by user while plug with the functin's ordinate.

*/
public class IntFloatLogAdapter extends IntFloatBoundedRangeAdapter {
	
	private float base, factor, offset;
	
    public IntFloatLogAdapter(DefaultFloatBoundedRangeModel model, int grades, float base, float factor, float offset) {
	    super(model, grades);
	    this.base = base;
	    this.factor = factor;
	    this.offset = offset;
    }
	
	protected float translate(int value) {
	    //System.out.println(offset + " + " + factor + " * log(" + base + ", " + value + ") = " + (offset + factor * (float) Math.pow(base, value)));
		return offset + factor * (float) Math.pow(base, value);
	}
	
	protected int translate(float value) {
		return (int) Math.round(Utils.log(base, (value - offset) / factor));
	}
	
}